#!/usr/bin/env python

"""...

"""

from __future__ import annotations
# from typing import Optional
import itertools

import numpy as np
import toytree
import toyplot
from toyplot.mark import Mark
from toyplot.coordinates import Axis
# from toyplot.html import RenderContext


class LabelHelper:
    """Controls the appearance and behavior of a Cartesian coordinate system label."""
    def __init__(self, text, style):
        self._style = {}
        self._text = None

        self.style = {
            "font-size": "14px",
            "font-weight": "bold",
            "stroke": "none",
            "text-anchor": "middle",
            "-toyplot-vertical-align": "bottom",
        }
        self.style = style
        self.text = text
        self.offset = 8


class Cart:
    """...

    Note
    ----
    See constructor function `canvas.cartesian` for user-facing func
    to create Cartesian axes.
    """
    def __init__(
        self,
        aspect: str = None,
        hyperlink: str = None,
        label: str = None,
        padding: int = 10,
        palette: toyplot.color.Palette = None,
        scenegraph: toyplot.scenegraph.SceneGraph = None,
        show: bool = True,
        xaxis: Axis = None,
        xlabel: str = None,
        xmax: float = None,
        xmax_range: float = None,
        xmin: float = None,
        xmin_range: float = None,
        xscale: str = "linear",
        xshow: bool = True,
        xticklocator: toyplot.locator.TickLocator = None,
        yaxis: Axis = None,
        ylabel: str = None,
        ymax: float = None,
        ymax_range: float = None,
        ymin: float = None,
        ymin_range: float = None,
        yscale: str = "linear",
        yshow: bool = True,
        yticklocator: toyplot.locator.TickLocator = None,
    ):
        self._finalized = Cart
        """: This Cartesian object after ._finalize() has been run on it."""
        self._scenegraph = scenegraph
        """: Graph connecting Canvas, Cartesian, and Marks. From Canvas."""

        self._xmin_range: float = xmin_range
        """: min range of data given Canvas dimensions and margin."""
        self._xmax_range: float = xmax_range
        """: max range of data given Canvas dimensions and margin."""
        self._ymin_range: float = ymin_range
        """: min range of data given Canvas dimensions and margin."""
        self._ymax_range: float = ymax_range
        """: max range of data given Canvas dimensions and margin."""

        self._aspect = aspect
        """: aspect options are 'fit-range' or None."""
        self._hyperlink = hyperlink
        """: str URI for axis ..."""
        self._label = LabelHelper(text=label, style={})
        """: Axes main label object w/ .text and .style attrs."""
        self._padding = toyplot.units.convert(padding, target="px", default="px")
        """: Padding between axes and data stored in px units."""
        self._palette = toyplot.color.Palette() if palette is None else palette
        """: Palette of colors used for sequentially added Marks."""

        # separate cycling palettes for different Mark types
        self._bar_colors = itertools.cycle(self._palette)
        self._ellipse_colors = itertools.cycle(self._palette)
        self._fill_colors = itertools.cycle(self._palette)
        self._graph_colors = itertools.cycle(self._palette)
        self._plot_colors = itertools.cycle(self._palette)
        self._scatterplot_colors = itertools.cycle(self._palette)
        self._rect_colors = itertools.cycle(self._palette)
        self._text_colors = itertools.cycle(self._palette)

        # set args to x Axis object
        self.x = xaxis if xaxis is not None else Axis()
        """: x Axis object for horizontal dimension of Cartesian."""
        self.x.show = xshow
        self.x.label.text = xlabel
        self.x.domain.min = xmin
        self.x.domain.max = xmax
        self.x.ticks.locator = xticklocator
        self.x.scale = xscale

        # set args to y Axis object
        self.y = yaxis if xaxis is not None else Axis()
        """: y Axis object for vertical dimension of Cartesian."""
        self.y.show = yshow
        self.y.label.text = ylabel
        self.y.domain.min = ymin
        self.y.domain.max = ymax
        self.y.ticks.locator = yticklocator
        self.y.scale = yscale

        # private attrs used in .finalize() to fit data & text extents
        self._expand_domain_range_x: np.ndarray = None
        """: array shape=(3, nMarks) w/ [position, left-extent, right-extent]"""
        self._expand_domain_range_y: np.ndarray = None
        """: array shape=(3, nMarks) w/ [position, top-extent, bottom-extent]"""

    # exposed property for set/get padding in px units
    @property
    def padding(self) -> float:
        return self._padding

    @padding.setter
    def padding(self, value: float) -> None:
        self._padding = toyplot.units.convert(value, target="px", default="px")

    # exposed property for set/get show boolean
    @property
    def show(self) -> bool:
        return self._show

    @show.setter
    def show(self, value: bool) -> None:
        self._show = value

    # exposed property for set/get hyperlink URI
    @property
    def hyperlink(self) -> str:
        return self._hyperlink

    @hyperlink.setter
    def hyperlink(self, value: str) -> None:
        self._hyperlink = value

    # exposed property for set/get aspect str
    @property
    def aspect(self) -> str:
        return self._aspect

    @aspect.setter
    def aspect(self, value: str) -> None:
        if value not in [None, "fit-range"]:
            raise ValueError(
                f"Unknown aspect value: {value}. Options are 'fit-range' or None.")
        self._aspect = value

    # private functions to set/get xmin_range used internally here.
    @property
    def xmin_range(self) -> float:
        return self._xmin_range

    @property
    def xmax_range(self) -> float:
        return self._xmax_range

    @property
    def ymin_range(self) -> float:
        return self._ymin_range

    @property
    def ymax_range(self) -> float:
        return self._ymax_range

    @xmin_range.setter
    def xmin_range(self, value: float) -> None:
        self._xmin_range = value

    @xmax_range.setter
    def xmax_range(self, value: float) -> None:
        self._xmax_range = value

    @ymin_range.setter
    def ymin_range(self, value: float) -> None:
        self._ymin_range = value

    @ymax_range.setter
    def ymax_range(self, value: float) -> None:
        self._ymax_range = value

    def project(self, axis, values) -> np.ndarray:
        pass

    def add_mark(self, mark):
        """Add a mark to the axes.

        This is only of use when creating your own custom Toyplot marks.  It is
        not intended for end-users.

        Example
        -------
        >>> mark = axes.add(MyCustomMark())

        Parameters
        ----------
        mark: :class:`toyplot.mark.Mark`, required

        Returns
        -------
        mark: :class:`toyplot.mark.Mark`
        """
        self._scenegraph.add_edge(self, "render", mark)
        self._scenegraph.add_edge(self.x, "map", mark)
        self._scenegraph.add_edge(self.y, "map", mark)
        return mark

    def share(self):
        pass

    # ----------------------------------------------------------------
    # Plotting functions that can be called from a Cart to create Marks
    # are located in cartesian_funcs.py and wrapped using functools.
    # ----------------------------------------------------------------
    # bars()
    # color_scale()
    # ellipse()
    # fill()
    # graph()
    # hlines()
    # plot()
    # rectangle()
    # scatterplot()
    # text()
    # vlines()
    # ----------------------------------------------------------------

    def _finalize(self) -> None:
        """...

        """
        if self._finalized is not None:
            return self._finalized

        # Begin with the implicit domain defined by our children.
        for name, axis in zip(("x", "y"), (self.x, self.y)):
            expand = []

            # iterate over Marks mapped to this axis
            for child in self._scenegraph.targets(axis, "map"):

                # get finalized version of Mark
                child = child._finalize()
                if child is not None:

                    # update display to fit Mark, and data if not annotation
                    axis._update_domain(
                        child.domain(name),
                        display=True,               # updates axis._display_[min,max]
                        data=not child.annotation,  # updates axis._data_[min,max]
                    )

                    # get coordinates and extents of child marks
                    (x, y), (left, right, top, bottom) = child.extents(["x", "y"])

                    # expand domain x range for mark positions
                    if name == "x":
                        expand.append([x, left, right])
                    else:
                        expand.append([y, top, bottom])

            # store x position and extents of Marks as arrays
            if expand:
                if name == "x":
                    self._expand_domain_range_x = np.concatenate(expand, axis=1)
                else:
                    self._expand_domain_range_y = np.concatenate(expand, axis=1)

        return self._finalized


if __name__ == "__main__":

    # from toytree.drawing.src.cartesian import Cart
    canvas = toyplot.Canvas()

    dims = ["xmin_range", "xmax_range", "ymin_range", "ymax_range"]
    layout = toyplot.layout.region(0, canvas._width, 0, canvas._height)
    kwargs = dict(zip(dims, layout))
    axes = Cart(scenegraph=canvas._scenegraph, **kwargs)
    canvas._scenegraph.add_edge(canvas, "render", axes)

    m0 = axes.text(0, 0, "hello world")
    m1 = axes.text(1, 1, "hello world")

    print(axes._expand_domain_range_x)
    print(axes._expand_domain_range_y)


#!/usr/bin/env python

"""
New Drawing class to create new mark and style on axes.
"""

# from copy import deepcopy, copy
from decimal import Decimal
import numpy as np
import toyplot

# from .Admixture import AdmixEdges

# for setting values from iterables
ITERABLE = (list, tuple, np.ndarray)



class GridSetup:
    """
    Returns Canvas and Cartesian axes objects to fit a grid of trees.
    """
    def __init__(self, nrows, ncols, width, height, layout):

        # style args can include height/width, nrows, ncols, shared,...
        self.nrows = nrows
        self.ncols = ncols
        self.width = width
        self.height = height
        self.layout = layout

        # get .canvas and .axes
        self.get_tree_dims()
        self.get_canvas_and_axes()


    def get_canvas_and_axes(self):
        """
        Set .canvas and .axes objects
        """
        self.canvas = toyplot.Canvas(
            height=self.height,
            width=self.width,
        )        

        self.axes = [
            self.canvas.cartesian(
                grid=(self.nrows, self.ncols, i),
                padding=10,
                margin=25,
            )
            for i in range(self.nrows * self.ncols)
        ]


    def get_tree_dims(self):
        """
        get height and width if not set by user
        """
        if self.ncols * self.nrows < 4:
            minx = 250
            miny = 250
        else:
            minx = 200
            miny = 140

        # wider than tall
        if self.layout in ("d", "u"):
            self.width = (
                self.width if self.width
                else min(750, minx * self.ncols)
            )
            self.height = (
                self.height if self.height
                else min(750, miny * self.nrows)
            )

        else:
            self.height = (
                self.height if self.height 
                else min(750, minx * self.nrows)
            )
            self.width = (
                self.width if self.width
                else min(750, miny * self.ncols)
            )



class CanvasSetup:
    """
    Returns Canvas and Cartesian axes objects 
    """
    def __init__(self, tree, axes, style):

        # args includes axes
        self.tree = tree
        self.axes = axes
        self.style = style
        self.canvas = None
        self.external_axis = False

        # get the longest name for dimension fitting
        self.lname = 0
        if not all([i is None for i in self.style.tip_labels]):
            self.lname = max([len(str(i)) for i in self.style.tip_labels])

        # ntips and shape to fit with provided args
        self.get_dims_from_tree_size()

        # fills canvas and axes
        self.get_canvas_and_axes()

        # expand the domain/extents for the text
        # self.fit_tip_labels()

        # ticks for tree and scalebar
        self.add_axes_style()


    def get_dims_from_tree_size(self):
        """
        Calculate reasonable canvas height and width for tree given N tips
        """
        if self.style.layout == "c":
            radius = max(
                [0] + [i for i in [self.style.height, self.style.width] if i])
            if not radius:
                radius = 400
            self.style.width = self.style.height = radius
            return

        if self.style.layout in ("r", "l"):
            # height fit by tree size
            if not self.style.height:
                self.style.height = max(275, min(1000, 18 * self.tree.ntips))
            # width fit by name size
            if not self.style.width:
                self.style.width = max(250, min(500, 250 + 5 * self.lname))
        else:
            # height fit by name size
            if not self.style.height:
                self.style.height = max(250, min(500, 250 + 5 * self.lname))
            # width fit by tree size
            if not self.style.width:
                self.style.width = max(350, min(1000, 18 * self.tree.ntips))



    def get_canvas_and_axes(self):
        """

        """
        if self.axes is not None: 
            self.canvas = None
            self.external_axis = True
        else:
            self.canvas = toyplot.Canvas(
                height=self.style.height,
                width=self.style.width,
            )
            self.axes = self.canvas.cartesian(
                padding=self.style.padding
            )



    def add_axes_style(self):
        """

        """
        # style axes with padding and show axes
        self.axes.padding = self.style.padding

        if not self.external_axis:
            self.axes.show = True
            if not self.style.scalebar:
                self.axes.show = False

        # scalebar        
        if self.style.scalebar:
            if self.style.layout in ("r", "l"):
                nticks = max((3, np.floor(self.style.width / 100).astype(int)))
                self.axes.y.show = False
                self.axes.x.show = True
                self.axes.x.ticks.show = True

                # generate locations
                if self.style.use_edge_lengths:
                    th = self.tree.treenode.height
                else:
                    th = self.tree.treenode.get_farthest_leaf(True)[1] + 1
                if self.style.layout == "r":
                    top = self.style.xbaseline - th
                else:
                    top = self.style.xbaseline + th
                locs = np.linspace(self.style.xbaseline, top, nticks)

                # auto-formatter for axes ticks labels
                zer = abs(min(0, Decimal(locs[1]).adjusted()))
                fmt = "{:." + str(zer) + "f}"
                self.axes.x.ticks.locator = toyplot.locator.Explicit(
                    locations=locs,
                    labels=[fmt.format(i) for i in np.abs(locs)],
                    )

            elif self.style.layout in ("u", "d"):
                nticks = max((3, np.floor(self.style.height / 100).astype(int)))
                self.axes.x.show = False
                self.axes.y.show = True
                self.axes.y.ticks.show = True

                # generate locations
                if self.style.use_edge_lengths:
                    th = self.tree.treenode.height
                else:
                    th = self.tree.treenode.get_farthest_leaf(True)[1] + 1
                if self.style.layout == "d":
                    top = self.style.ybaseline + th
                else:
                    top = self.style.ybaseline - th
                locs = np.linspace(self.style.ybaseline, top, nticks)

                # auto-formatter for axes ticks labels
                zer = abs(min(0, Decimal(locs[1]).adjusted()))
                fmt = "{:." + str(zer) + "f}"
                self.axes.y.ticks.locator = toyplot.locator.Explicit(
                    locations=locs,
                    labels=[fmt.format(i) for i in np.abs(locs)],
                    )

            # elif self.style.layout == "d":
            #     nticks = max((3, np.floor(self.style.height / 100).astype(int)))
            #     self.axes.x.show = False
            #     self.axes.y.show = True
            #     self.axes.y.ticks.show = True            

            #     # generate locations
            #     locs = np.linspace(0, self.tree.treenode.height, nticks)

            #     # auto-formatter for axes ticks labels
            #     zer = abs(min(0, Decimal(locs[1]).adjusted()))
            #     fmt = "{:." + str(zer) + "f}"
            #     self.axes.y.ticks.locator = toyplot.locator.Explicit(
            #         locations=locs,
            #         labels=[fmt.format(i) for i in np.abs(locs)],
            #         )



    # def fit_tip_labels(self):
    #     """
    #     DEPRECATED SINCE V2 since Mark now sets its own extents correctly.

    #     Modifies display range to ensure tip labels fit. This is a bit hackish
    #     still. The problem is that the 'extents' range of the rendered text
    #     is not totally correct. So we add a little buffer here. Should add for 
    #     user to be able to modify this if needed. If not using edge lengths
    #     then need to use unit length for treeheight.
    #     """
    #     # bail on unrooted for now; TODO
    #     if self.style.layout == "c":
    #         return

    #     # if names
    #     if self.lname:
    #         # get ratio of names to tree in plot
    #         ratio = max(self.lname / 10, 0.15)

    #         # have tree figure make up 85% of plot
    #         if self.style.use_edge_lengths:
    #             addon = self.tree.treenode.height
    #         else:
    #             addon = self.tree.treenode.get_farthest_leaf(True)[1] + 1
    #         addon *= ratio

    #         # modify display for layout
    #         if self.style.layout == "r":
    #             self.axes.x.domain.max = (addon / 2.) + self.style.xbaseline
    #         elif self.style.layout == "l":
    #             self.axes.x.domain.min = (-addon / 2.) + self.style.xbaseline
    #             # self.axes.x.domain.min -= self.style.xbaseline
    #         elif self.style.layout == "d":
    #             self.axes.y.domain.min = (-addon / 2.) + self.style.ybaseline
    #         elif self.style.layout == "u":
    #             self.axes.y.domain.max = (addon / 2.) + self.style.ybaseline

    #         # print(addon, ratio, self.axes.x.domain.min, self.axes.x.domain.max)



#!/usr/bin/env python

"""Methods of the Cartesian class to create Marks.

In contrast to toyplot the methods of Cartesian that users can
call to create Marks are defined here in a separate module and
mapped to the Cartesian class using an API.
"""

from typing import Sequence, Mapping, Any, TypeVar
from toyplot.mark import Mark
import toyplot
from toytree.utils.src.toyplotfix.cartesian import Cartesian
from toyplot.cartesian_api import add_cartesian_method

Color = TypeVar("Color")

# list of methods to export
__all__ = ["text",]


def _mark_exportable(table, column, exportable=True):
    table.metadata(column)["toyplot:exportable"] = exportable


@add_cartesian_method(Cartesian)
def text(
    axes: Cartesian,
    a: Sequence[float],
    b: Sequence[float],
    text: Sequence[str],
    angle: float = 0,
    color: Color = None,
    opacity: float = 1.0,
    title: str = None,
    style: Mapping[str, Any] = None,
    filename: str = None,
    annotation: bool = True,
) -> Mark:
    """Add text to the axes.

    Parameters
    ----------
    a, b: float
      Coordinates of the text anchor.
    text: string
      The text to be displayed.
    title: string, optional
      Human-readable title for the mark.  The SVG / HTML backends render the
      title as a tooltip.
    style: dict, optional
      Collection of CSS styles to apply to the mark.  See
      :class:`toyplot.mark.Text` for a list of useful styles.
    annotation: boolean, optional
      Set to True if this mark should be considered an annotation.

    Returns
    -------
    text: :class:`toyplot.mark.Text`
    """
    table = toyplot.data.Table()
    table["x"] = toyplot.require.scalar_vector(a)
    table["y"] = toyplot.require.scalar_vector(b, table.shape[0])
    table["text"] = toyplot.broadcast.pyobject(text, table.shape[0])
    _mark_exportable(table, "x")
    _mark_exportable(table, "y")
    _mark_exportable(table, "text")

    table["angle"] = toyplot.broadcast.scalar(angle, table.shape[0])
    table["opacity"] = toyplot.broadcast.scalar(opacity, table.shape[0])
    table["title"] = toyplot.broadcast.pyobject(title, table.shape[0])
    style = toyplot.style.require(style, allowed=toyplot.style.allowed.text)

    default_color = [next(axes._text_colors)]

    color = toyplot.color.broadcast(
        colors=color,
        shape=(table.shape[0], 1),
        default=default_color,
    )
    table["fill"] = color[:, 0]

    mark = toyplot.mark.Text(
        coordinate_axes=["x", "y"],
        table=table,
        coordinates=["x", "y"],
        text=["text"],
        angle=["angle"],
        fill=["fill"],
        opacity=["opacity"],
        title=["title"],
        style=style,
        annotation=annotation,
        filename=filename,
    )
    return axes.add_mark(mark)


if __name__ == "__main__":

    pass
    # canvas = toyplot.Canvas()
    # axes = toyplot.cartesian2.Cartesian2(scenegraph=canvas._scenegraph)
    # canvas._scenegraph.add_edge(canvas, "render", axes)

    # print(dir(axes))
    # m0 = axes.text2(0, 0, "hello world")
    # m1 = axes.text2(1, 1, "hello world")

    # print(m0.extents('x'))
    # # toytree.utils.show(canvas)



#!/usr/bin/env python

"""Cartesian class object.

In toyplot the Cartesian class is defined in coordinates.py with many
other objects and methods. Here it is isolated and simplified for
easier debugging.

TODO
----
- break finalize into several smaller functions.
- copy or inherit render function from toyplot.coordinates.Cartesian

"""

from __future__ import annotations
from typing import Optional, Sequence, Mapping, Any
from dataclasses import dataclass
import itertools
from loguru import logger

import numpy as np
import toyplot
import toyplot.text
from toyplot.mark import Mark
from toyplot.axis import Axis, LabelHelper


@dataclass(repr=False)
class CartesianLabelHelper(LabelHelper):
    """Controls the appearance and behavior of an Cartesian label."""
    def __post_init__(self):
        self.style = {
            "font-size": "14px",
            "font-weight": "bold",
            "stroke": "none",
            "text-anchor": "middle",
            "-toyplot-vertical-align": "bottom",
        }
        self.offset = 8


class Cartesian(toyplot.coordinates.Cartesian):
    """Cartesian class to map Marks on cartesian axes.

    """
    def __init__(
        self,
        aspect: str = None,
        hyperlink: str = None,
        label: str = None,
        padding: int = 10,
        palette: toyplot.color.Palette = None,
        scenegraph: toyplot.scenegraph.SceneGraph = None,
        show: bool = True,
        xaxis: Axis = None,
        xlabel: str = None,
        xmax: float = None,
        xmax_range: float = None,
        xmin: float = None,
        xmin_range: float = None,
        xscale: str = "linear",
        xshow: bool = True,
        xticklocator: toyplot.locator.TickLocator = None,
        yaxis: Axis = None,
        ylabel: str = None,
        ymax: float = None,
        ymax_range: float = None,
        ymin: float = None,
        ymin_range: float = None,
        yscale: str = "linear",
        yshow: bool = True,
        yticklocator: toyplot.locator.TickLocator = None,
    ):
        self._finalized = None
        """: This Cartesian object after ._finalize() has been run on it."""
        self._scenegraph = scenegraph
        """: Graph connecting Canvas, Cartesian, and Marks. From Canvas."""

        self._xmin_range: float = xmin_range
        """: min range of data given Canvas dimensions and margin."""
        self._xmax_range: float = xmax_range
        """: max range of data given Canvas dimensions and margin."""
        self._ymin_range: float = ymin_range
        """: min range of data given Canvas dimensions and margin."""
        self._ymax_range: float = ymax_range
        """: max range of data given Canvas dimensions and margin."""

        self._show = show
        """: Show or suppress both Cartesian axes."""
        self._aspect = aspect
        """: aspect options are 'fit-range' or None."""
        self._hyperlink = hyperlink
        """: str URI for axis ..."""
        self._label = CartesianLabelHelper(_text=label, _style={})
        """: Axes main label object w/ .text and .style attrs."""
        self._padding = toyplot.units.convert(padding, target="px", default="px")
        """: Padding between axes and data stored in px units."""
        self._palette = toyplot.color.Palette() if palette is None else palette
        """: Palette of colors used for sequentially added Marks."""

        # distinct cycling palettes for different Mark types
        self._bar_colors = itertools.cycle(self._palette)
        self._ellipse_colors = itertools.cycle(self._palette)
        self._fill_colors = itertools.cycle(self._palette)
        self._graph_colors = itertools.cycle(self._palette)
        self._plot_colors = itertools.cycle(self._palette)
        self._scatterplot_colors = itertools.cycle(self._palette)
        self._rect_colors = itertools.cycle(self._palette)
        self._text_colors = itertools.cycle(self._palette)

        # set args to x Axis object
        self.x = xaxis if xaxis is not None else Axis()
        """: x Axis object for horizontal dimension of Cartesian."""
        self.x.show = xshow
        self.x.label.text = xlabel
        self.x.domain.min = xmin
        self.x.domain.max = xmax
        self.x.ticks.locator = xticklocator
        self.x.scale = xscale

        # set args to y Axis object
        self.y = yaxis if xaxis is not None else Axis()
        """: y Axis object for vertical dimension of Cartesian."""
        self.y.show = yshow
        self.y.label.text = ylabel
        self.y.domain.min = ymin
        self.y.domain.max = ymax
        self.y.ticks.locator = yticklocator
        self.y.scale = yscale

        # private attrs used in .finalize() to fit data & text extents
        self._expand_domain_range_x: np.ndarray = None
        """: array shape=(3, nMarks) w/ [position, left-extent, right-extent]"""
        self._expand_domain_range_y: np.ndarray = None
        """: array shape=(3, nMarks) w/ [position, top-extent, bottom-extent]"""

    # exposed property for set/get padding in px units
    @property
    def padding(self) -> float:
        return self._padding

    @padding.setter
    def padding(self, value: float) -> None:
        self._padding = toyplot.units.convert(value, target="px", default="px")

    # exposed property for set/get show boolean
    @property
    def show(self) -> bool:
        return self._show

    @show.setter
    def show(self, value: bool) -> None:
        self._show = value

    # exposed property for set/get hyperlink URI
    @property
    def hyperlink(self) -> str:
        return self._hyperlink

    @hyperlink.setter
    def hyperlink(self, value: str) -> None:
        self._hyperlink = value

    # exposed property for set/get aspect str
    @property
    def aspect(self) -> str:
        return self._aspect

    @aspect.setter
    def aspect(self, value: str) -> None:
        if value not in [None, "fit-range"]:
            raise ValueError(
                f"Unknown aspect value: {value}. Options are 'fit-range' or None.")
        self._aspect = value

    # private functions to set/get xmin_range used internally here.
    @property
    def xmin_range(self) -> float:
        return self._xmin_range

    @property
    def xmax_range(self) -> float:
        return self._xmax_range

    @property
    def ymin_range(self) -> float:
        return self._ymin_range

    @property
    def ymax_range(self) -> float:
        return self._ymax_range

    @xmin_range.setter
    def xmin_range(self, value: float) -> None:
        self._xmin_range = value

    @xmax_range.setter
    def xmax_range(self, value: float) -> None:
        self._xmax_range = value

    @ymin_range.setter
    def ymin_range(self, value: float) -> None:
        self._ymin_range = value

    @ymax_range.setter
    def ymax_range(self, value: float) -> None:
        self._ymax_range = value

    def project(self, axis: str, values: np.ndarray) -> np.ndarray:
        """Project a set of domain values to coordinate system range values.

        Note that this API is intended for advanced users creating their own
        custom marks, end-users should never need to use it.

        Parameters
        ----------
        axis: "x" or "y", required
            The axis to be projected
        values: array-like, required
            The values to be projected

        Returns
        -------
        projected: :class:`numpy.ndarray`
            The projected values.
        """
        if axis == "x":
            return self._x_projection(values)
        elif axis == "y":
            return self._y_projection(values)
        raise ValueError("Unexpected axis: %s" % axis)

    def add_mark(self, mark: Mark) -> Mark:
        """Add a mark to the axes.

        This is only of use when creating your own custom Toyplot marks.  It is
        not intended for end-users.

        Example
        -------
        >>> mark = axes.add(MyCustomMark())

        Parameters
        ----------
        mark: :class:`toyplot.mark.Mark`, required

        Returns
        -------
        mark: :class:`toyplot.mark.Mark`
        """
        self._scenegraph.add_edge(self, "render", mark)
        self._scenegraph.add_edge(self.x, "map", mark)
        self._scenegraph.add_edge(self.y, "map", mark)
        return mark

    def share(self):
        """TODO later."""
        pass

    # ----------------------------------------------------------------
    # Plotting functions that can be called from a Cart to create Marks
    # are located in cartesian2_funcs.py and wrapped using functools.
    # ----------------------------------------------------------------
    # bars()
    # color_scale()
    # ellipse()
    # fill()
    # graph()
    # hlines()
    # plot()
    # rectangle()
    # scatterplot()
    # text()
    # vlines()
    # ----------------------------------------------------------------

    def _finalize(self) -> None:
        """Update domain and range to fit Marks mapped to Cartesian axes."""
        # if already finalized then skip.
        if self._finalized is not None:
            return self._finalized

        # get Marks mapped to 'x' or 'y' (all are mapped to both for Cartesian)
        marks = self._scenegraph.targets(self.x, "map")

        # dict to store extents in each direction
        keys = ("x", "y", "left", "right", "top", "bottom")
        expand = {i: [] for i in keys}

        # iterate over marks updating domain and getting extents
        for mark in marks:

            # get finalized version of Mark
            fmark = mark._finalize()
            if fmark is not None:

                # get coordinates and extents of child marks
                (x, y), (left, right, top, bottom) = mark.extents(["x", "y"])

                # updates Axis._display_[min,max] and Axis._data_[min,max]
                # to fit Mark. Data only updated if not annotation.
                for ax in ('x', 'y'):
                    getattr(self, ax)._update_domain(
                        mark.domain(ax),
                        display=True,
                        data=not mark.annotation
                    )

                # store info to expand range to fit extents
                for key, val in zip(keys, (x, y, left, right, top, bottom)):
                    expand[key].extend(val)

        # convert expand values into arrays and store to self, else None
        for key, val in expand.items():
            # if not val:
            #     val = None
            attr = f"_expand_domain_range_{key}"
            setattr(self, attr, np.array(val))

            # extents...
            logger.info(f"{key}, {val}")

        # get implicit domain from data (data units), or use origin.
        xdomain_min = self.x._display_min if self.x._display_min else 0
        xdomain_max = self.x._display_max if self.x._display_max else 0
        ydomain_min = self.y._display_min if self.y._display_min else 0
        ydomain_max = self.y._display_max if self.y._display_max else 0

        # ensure domain is not empty (e.g., min and max 0).
        if xdomain_min == xdomain_max:
            xdomain_min -= 0.5
            xdomain_max += 0.5
        if ydomain_min == ydomain_max:
            ydomain_min -= 0.5
            ydomain_max += 0.5

        # ----------------- canvas width --------------------------
        # ------- self._xmin_range
        #                                  self._xmax_range -------
        #              |----------- range -----------|
        # margin + pad + ext-left + data + ext-right + pad + margin
        # note: margin left does affect _xmax_range, only margin right

        # get projection function to translate current domain in data
        # units to range in canvas (e.g., px) units. Example, project
        # data domain (0, 1) to range (50, 350) on a Canvas with
        # width=400 and margin=50. Padding is not relevant here.
        self.x_projection = _create_projection(
            self.x.scale,
            domain_min=xdomain_min,
            domain_max=xdomain_max,
            range_min=self._xmin_range,
            range_max=self._xmax_range,
        )
        logger.info(f"orig domain = {xdomain_min}, {xdomain_max}")
        logger.info(f"orig data domain to px: {self.x_projection([xdomain_min, xdomain_max])}")

        # get range needed to fit marks on the new expanded x range
        range_x = self.x_projection(self._expand_domain_range_x)
        logger.info(f"range_x: {range_x}")#self.x_projection([xdomain_min, xdomain_max])}")

        # get range left and right of marks needed to fit their extents
        range_left = range_x + self._expand_domain_range_left
        range_right = range_x + self._expand_domain_range_right
        logger.info(f"ext range: {range_x}, left: {range_left.round(3)}, right: {range_right.round(3)}")

        # project left and right from canvas units back into data units
        # and get the min max values to set as new domain min max.
        # e.g., prior domain was (0, 1) and now it may be (0.1, 0.9).
        domain_left = self.x_projection.inverse(range_left)
        domain_right = self.x_projection.inverse(range_right)
        xdomain_min, xdomain_max = toyplot.data.minimax(
            [xdomain_min, xdomain_max, domain_left, domain_right])
        logger.info(f"re domain = {xdomain_min}, {xdomain_max}")
        # min_range = min(range_left)
        # max_range = max(range_right)
        # logger.warning([min_range, max_range])
        # get min and max of range to fit extents
        # eidx = np.argmax(range_right)
        # right_most_extent_width = self._expand_domain_range_right[eidx]
        # eidx = np.argmin(range_left)
        # left_most_extent_width = self._expand_domain_range_left[eidx]
        # new_min_range = self._xmin_range - left_most_extent_width
        # new_max_range = self._xmax_range - right_most_extent_width

        # extents cannot push right beyond limits
        # print(new_min_range, new_max_range)
        # if new_min_range > new_max_range:
        #     mid = (new_min_range + new_max_range) / 2.
        #     new_min_range = mid - 1e-6
        #     new_max_range = mid + 1e-6

        # get projection function to translate current domain to expanded range
        # x_projection = _create_projection(
        #     self.x.scale,
        #     domain_min=xdomain_min,
        #     domain_max=xdomain_max,
        #     range_min=min_range,#20 + abs(left_most_extent_width), #new_min_range,
        #     range_max=max_range,##345.8 - 40,#new_max_range,
        # )

        # domain_left = x_projection.inverse(range_left)
        # domain_right = x_projection.inverse(range_right)
        # x0, x1 = x_projection.inverse([20, 460])
        # print('x0', x0, x1)
        # print("d", domain_left, domain_right)

        # # update domain
        # xdomain_min, xdomain_max = toyplot.data.minimax(
        #     [xdomain_min, xdomain_max, domain_left, domain_right])
        # print("xd", xdomain_min, xdomain_max)

        # ...
        self.y_projection = _create_projection(
            self.y.scale,
            domain_min=ydomain_min,
            domain_max=ydomain_max,
            range_min=self._ymin_range,
            range_max=self._ymax_range,
        )
        logger.info(f"orig domain = {ydomain_min}, {ydomain_max}")
        range_y = self.y_projection(self._expand_domain_range_y)
        range_top = range_y + self._expand_domain_range_top
        range_bottom = range_y + self._expand_domain_range_bottom
        domain_top = self.y_projection.inverse(range_top)
        domain_bottom = self.y_projection.inverse(range_bottom)
        ydomain_min, ydomain_max = toyplot.data.minimax(
            [ydomain_min, ydomain_max, domain_top, domain_bottom])
        logger.info(f"re domain = {ydomain_min}, {ydomain_max}")

        # optionally expand the domain to match x and y ranges
        if self._aspect == "fit-range":
            dwidth = xdomain_max - xdomain_min
            dheight = ydomain_max - ydomain_min
            daspect = dwidth / dheight
            raspect = (
                (self._xmax_range - self._xmin_range) /
                (self._ymax_range - self._ymin_range)
            )
            if daspect < raspect:
                offset = ((dwidth * (raspect / daspect)) - dwidth) * 0.5
                xdomain_min -= offset
                xdomain_max += offset
            elif daspect > raspect:
                offset = ((dheight * (daspect / raspect)) - dheight) * 0.5
                ydomain_min -= offset
                ydomain_max += offset

        # allow users to override the domain
        xdomain_min = self.x.domain.min if xdomain_min is None else xdomain_min
        xdomain_max = self.x.domain.max if xdomain_max is None else xdomain_max
        ydomain_min = self.y.domain.min if ydomain_min is None else ydomain_min
        ydomain_max = self.y.domain.max if ydomain_max is None else ydomain_max

        # ensure the domain is not empty after user could have modified
        if xdomain_min == xdomain_max:
            xdomain_min -= 0.5
            xdomain_max += 0.5
        if ydomain_min == ydomain_max:
            ydomain_min -= 0.5
            ydomain_max += 0.5

        # calculate x tick locations and labels
        xtick_locations = []
        xtick_labels = []
        xtick_titles = []
        if self.show and self.x.show:
            loc = self.x._locator()
            xtick_locations, xtick_labels, xtick_titles = (
                loc.ticks(xdomain_min, xdomain_max))

        # calculate y tick locations and labels
        ytick_locations = []
        ytick_labels = []
        ytick_titles = []
        if self.show and self.y.show:
            loc = self.y._locator()
            ytick_locations, ytick_labels, ytick_titles = (
                loc.ticks(ydomain_min, ydomain_max))

        # Allow tick locations to grow (never shrink) the domain.
        if len(xtick_locations):
            xdomain_min = np.amin((xdomain_min, xtick_locations[0]))
            xdomain_max = np.amax((xdomain_max, xtick_locations[-1]))
        if len(ytick_locations):
            ydomain_min = np.amin((ydomain_min, ytick_locations[0]))
            ydomain_max = np.amax((ydomain_max, ytick_locations[-1]))

        # Create projections for each axis.
        self._x_projection = _create_projection(
            scale=self.x.scale,
            domain_min=xdomain_min,
            domain_max=xdomain_max,
            range_min=self._xmin_range,
            range_max=self._xmax_range,
        )
        self._y_projection = _create_projection(
            scale=self.y.scale,
            domain_min=ydomain_min,
            domain_max=ydomain_max,
            range_min=self._ymax_range,
            range_max=self._ymin_range,
        )

        # Finalize positions for all axis components.
        if self.x.spine.position == "low":
            x_offset = self.padding
            x_spine_y = self._ymax_range
            x_ticks_near = 0
            x_ticks_far = 5
            x_tick_location = "below"
            x_label_location = "below"
        elif self.x.spine.position == "high":
            x_offset = -self.padding  # pylint: disable=invalid-unary-operand-type
            x_spine_y = self._ymin_range
            x_ticks_near = 5
            x_ticks_far = 0
            x_tick_location = "above"
            x_label_location = "above"
        else:
            x_offset = 0
            x_spine_y = self._y_projection(self.x.spine.position)
            x_ticks_near = 3
            x_ticks_far = 3
            x_tick_location = "below"
            x_label_location = "below"

        if self.y.spine._position == "low":
            y_offset = -self.padding  # pylint: disable=invalid-unary-operand-type
            y_spine_x = self._xmin_range
            y_ticks_near = 0
            y_ticks_far = 5
            y_tick_location = "above"
            y_label_location = "above"
        elif self.y.spine._position == "high":
            y_offset = self.padding
            y_spine_x = self._xmax_range
            y_ticks_near = 0
            y_ticks_far = 5
            y_tick_location = "below"
            y_label_location = "below"
        else:
            y_offset = 0
            y_spine_x = self._x_projection(self.y.spine._position)
            y_ticks_near = 3
            y_ticks_far = 3
            y_tick_location = "below"
            y_label_location = "below"

        # Finalize the axes.
        self.x._finalize(
            x1=self._xmin_range,
            x2=self._xmax_range,
            y1=x_spine_y,
            y2=x_spine_y,
            offset=x_offset,
            domain_min=xdomain_min,
            domain_max=xdomain_max,
            tick_locations=xtick_locations,
            tick_labels=xtick_labels,
            tick_titles=xtick_titles,
            default_tick_location=x_tick_location,
            default_ticks_far=x_ticks_far,
            default_ticks_near=x_ticks_near,
            default_label_location=x_label_location,
        )
        self.y._finalize(
            x1=y_spine_x,
            x2=y_spine_x,
            y1=self._ymax_range,
            y2=self._ymin_range,
            offset=y_offset,
            domain_min=ydomain_min,
            domain_max=ydomain_max,
            tick_locations=ytick_locations,
            tick_labels=ytick_labels,
            tick_titles=ytick_titles,
            default_tick_location=y_tick_location,
            default_ticks_far=y_ticks_far,
            default_ticks_near=y_ticks_near,
            default_label_location=y_label_location,
        )
        self._finalized = self
        return self._finalized


def _create_projection(scale, domain_min, domain_max, range_min, range_max):
    if isinstance(scale, toyplot.projection.Projection):
        return scale
    if scale == "linear":
        return toyplot.projection.linear(domain_min, domain_max, range_min, range_max)
    scale, base = scale
    return toyplot.projection.log(base, domain_min, domain_max, range_min, range_max)


def _mark_exportable(table, column, exportable=True):
    table.metadata(column)["toyplot:exportable"] = exportable



if __name__ == "__main__":

    import toyplot.browser
    import toyplot.dispatch
    canvas = toyplot.Canvas(width=500, height=500)

    dims = ["xmin_range", "xmax_range", "ymin_range", "ymax_range"]
    layout = toyplot.layout.region(0, canvas._width, 0, canvas._height)
    kwargs = dict(zip(dims, layout))
    axes = Cartesian(scenegraph=canvas._scenegraph, **kwargs)
    canvas._scenegraph.add_edge(canvas, "render", axes)

    # m0 = axes.text2(0, 0, "hello world")
    # m2 = axes.text2(0.5, 0.5, "hello world")
    # m2 = axes.text2(0.75, 0.75, "hello world")
    # m1 = axes.text2(1, 1, "hello world")

    # axes._finalize()
    # print(axes._expand_domain_range_right)
    # print(axes._expand_domain_range_y)
    # print()
    toyplot.browser.show(canvas)



#!/usr/bin/env python

import itertools
import toyplot
import numpy as np
from .Render import split_rgba_style
from .NodeAssist import NodeAssist
from .TreeStyle import COLORS1
from .utils import ToytreeError

ITERABLE = (list, tuple, np.ndarray)

"""
Checks style dictionaries for types and size and builds out arguments
into arrays for variable styles.
"""


class StyleChecker:
    """
    Checks for allowed styles in style dictionaries and expands args
    for individuals styles into arrays while checking types.
    """
    def __init__(self, ttree, style):

        # input objects
        self.ttree = ttree
        self.style = style

        # dimensions
        self.ntips = self.ttree.ntips
        self.nnodes = self.ttree.nnodes
        self.nedges = self.ttree._coords.edges.shape[0]

        # check dictionary inputs 
        self.check_dicts()

        # expand to arrays and sometimes set 'fill' or 'stroke' in dicts.
        self.expand_vars()



    def check_dicts(self):
        """
        Checks style dicts for allowable styles. 
        TODO: check for None and convert to "none"?
        TODO: convert all fill/stroke to rgb and stroke-opacity.
        """

        # check for allowable types
        self.style.node_style = toyplot.style.require(
            self.style.node_style, toyplot.style.allowed.marker)
        self.style.edge_style = toyplot.style.require(
            self.style.edge_style, toyplot.style.allowed.line)
        self.style.edge_align_style = toyplot.style.require(
            self.style.edge_align_style, toyplot.style.allowed.line)
        self.style.tip_labels_style = toyplot.style.require(
            self.style.tip_labels_style, toyplot.style.allowed.text)
        self.style.node_labels_style = toyplot.style.require(
            self.style.node_labels_style, toyplot.style.allowed.text)

        sdicts = [
            self.style.node_style,
            self.style.edge_style,
            self.style.edge_align_style,
            self.style.tip_labels_style,
            self.style.node_labels_style,
        ]

        # check for None and split colors
        for sdict in sdicts:

            # next set the fill or stroke by splitting rgba to rgb and opacity
            # and set the -opacity unless the unser explicitly set it already.
            for coltype in ['fill', 'stroke']:
                if sdict.get(coltype):

                    # convert to css color if not already
                    color = sdict[coltype]

                    try:
                        color = toyplot.color.css(color)
                    except Exception:
                        pass

                    # split the color
                    coldict = split_rgba_style({coltype: color})
                    sdict[coltype] = coldict[coltype]

                    # set the opacity of this type unless already set
                    okey = "{}-opacity".format(coltype)
                    if not sdict.get(okey):
                        sdict[okey] = coldict[okey]




    def expand_vars(self):
        """
        Expand args for individual styles into arrays of correct length. If
        an indiv style was set with only one value then we instead set the 
        style dict fill or stroke (and opacity) with all properly using rgb
        and opacity settings.
        """
        self._assign_node_labels()
        self._assign_node_colors()
        self._assign_node_sizes()
        self._assign_node_shapes()
        self._assign_node_hover()

        self._assign_tip_colors()
        self._assign_tip_labels()
        self._assign_tip_labels_angles()

        self._assign_edge_colors()
        self._assign_edge_widths()

        self._assign_admixture_idxs()



    def _assign_tip_colors(self):
        """
        Sets .tip_colors as an array of length ntips. Only fills vals if there
        is variation among tips, otherwise leaves empty to use style fill.

        tip_colors=None       |  [None, None, ...]
        tip_colors=ITERABLE   |  [col0, col1, col2, ...] 
        tip_colors=STR        |  [None, None, ...] & tip_labels_style['fill']
        """
        arg = self.style.tip_labels_colors

        # set it all empty
        if arg is None:
            self.style.tip_labels_colors = toyplot.broadcast.pyobject(None, self.ntips)

        else:
            self.style.tip_labels_colors = toyplot.color.broadcast(arg, self.ntips, None)

            # if all the same then reset to None
            if len(set([str(i) for i in self.style.tip_labels_colors])) == 1:

                # save the fixed color and set to None
                color = self.style.tip_labels_colors[0]
                self.style.tip_labels_colors = toyplot.broadcast.pyobject(None, self.ntips)

                # set edge_style stroke and stroke-opacity 
                sub = split_rgba_style({'fill': color})
                self.style.tip_labels_style['fill'] = sub["fill"]
                self.style.tip_labels_style['fill-opacity'] = sub["fill-opacity"]

            else:
                # reorder array for fixed if needed
                if self.style.fixed_order:
                    self.style.tip_labels_colors = self.style.tip_labels_colors[self.ttree._fixed_idx]


    def _assign_tip_labels(self):
        """
        Sets .tip_labels as an array of length ntips.

        tip_labels=None             | ['a', 'b', 'c'] or fixed_order
        tip_labels=True             | ['a', 'b', 'c'] or fixed_order
        tip_labels=False            | [None, None, None]
        tip_labels=['a', 'c', 'b']  | ['a', 'c', 'b']
        """
        arg = self.style.tip_labels

        # set it
        if arg is False:
            self.style.tip_labels = None
        elif isinstance(arg, ITERABLE):
            pass
        else:  # arg in [None, True]:
            self.style.tip_labels = self.ttree.get_tip_labels()

        # check length
        self.style.tip_labels = toyplot.broadcast.pyobject(
            self.style.tip_labels, self.ntips)


    def _assign_tip_labels_angles(self):
        """
        Sets .tip_labels_angles as an array of length ntips. No user 
        args for this one, it inherits from the layout type.
        """
        # set circular layout
        if self.style.layout == 'c':

            # define range of tip radians
            tip_radians = np.linspace(0, -np.pi * 2, self.ntips + 1)[:-1]
            angles = np.array([np.rad2deg(abs(i)) for i in tip_radians]) * -1

        elif self.style.layout == "u":
            angles = -90

        elif self.style.layout == "d":
            angles = -90

        elif self.style.layout == "l":
            angles = 0

        else:
            angles = 0

        self.style.__dict__['tip_labels_angles'] = toyplot.broadcast.scalar(angles, self.ntips)


    def _assign_node_labels(self):
        """
        Sets .node_labels array of length nnodes. In addition to being the
        text labels on nodes these features can also affect whether node
        marks will be shown. Suppressed markers are recorded with nan as 
        the value label. To show the node mark but with empty labels a None
        value is used. 

        node_labels=None                        |  [None, None, None,]
        node_labels=False                       |  [None, None, None,]
        node_labels=True                        |  [0, 1, 2, 3]
        node_labels=ITERABLE                    |  [100, 90, 95, ...]
        node_labels=STR in .feat, e.g., "idx"   |  [nan, nan, 2, ..., nan]
        node_labels=TUPLE, e.g., ("idx", 1, 0)  |  [0, 1, 2, ..., nan]
        """
        arg = self.style.node_labels

        # set it.
        if arg is True:
            self.style.node_labels = self.ttree.get_node_values("idx", 1, 1)[::-1]
        elif arg is False:
            self.style.node_labels = None
        elif arg is None:
            self.style.node_labels = None
        elif isinstance(arg, (str, bytes)):
            if arg in self.ttree.features:
                self.style.node_labels = self.ttree.get_node_values(arg, 0, 0)[::-1]
        elif isinstance(arg, tuple) and (len(arg) == 3):
            if arg[0] in self.ttree.features:
                self.style.node_labels = self.ttree.get_node_values(*arg)[::-1]
        elif isinstance(arg, ITERABLE):
            self.style.node_labels = arg[::-1]
        else:
            raise ToytreeError(
                "node_labels argument not recognized: {}".format(arg))

        # check length
        self.style.node_labels = toyplot.broadcast.pyobject(
            self.style.node_labels, self.nnodes)


    def _assign_node_shapes(self):
        """
        Sets .node_shapes as an array of length nnodes in levelorder

        node_markers=None         | ['o', 'o', 'o', ...]
        node_markers=STR          | ['o', 'o', 'o', ...]
        node_markders=ITERABLE     | ['o', 's', 'o', ...]
        """
        arg = self.style.node_markers
        if isinstance(arg, ITERABLE):
            self.style.node_markers = arg[::-1]
        elif isinstance(arg, (bytes, str)):
            self.style.node_markers = arg
        else:
            self.style.node_markers = 'o'

        # project to size
        self.style.node_markers = toyplot.broadcast.pyobject(
            self.style.node_markers, self.nnodes)


    def _assign_node_sizes(self):
        """
        Sets .node_sizes as an array of length nnodes in levelorder

        node_sizes=None         | [None, None, None, ...]
        node_sizes=INT          | [12, 12, 12, 12, ...]
        node_sizes=ITERABLE     | [5, 12, 20, 5, ...]
        """
        # 
        arg = self.style.node_sizes
        if isinstance(arg, ITERABLE):
            self.style.node_sizes = arg[::-1]
        elif isinstance(arg, (int, float)):
            self.style.node_sizes = arg
        else:
            self.style.node_sizes = None

        # project to size
        self.style.node_sizes = toyplot.broadcast.pyobject(
            self.style.node_sizes, self.nnodes)

        # special: hide nodes with labels that are (TODO: nan) for now ("")
        # but only if arg was None or INT, not if explicity list type.
        if isinstance(arg, (int, float)):
            self.style.node_sizes[self.style.node_labels == ""] = 0


    def _assign_node_hover(self):
        """
        Sets .node_hover as an array of length nnodes in levelorder.

        node_hover=None         | [None, None, None, ...]
        node_hover=True         | [node_idx=31\nname=blah\nfeat=x, ...]
        node_hover="idx"        | [31, 30, 29, 28, ...]
        node_hover=ITERABLE     | ['a', 'b', 'c', 'd', ...]
        """
        # 
        arg = self.style.node_hover
        if isinstance(arg, ITERABLE):
            self.style.node_hover = arg[::-1]
        elif arg is None:
            self.style.node_hover = None
        elif arg is False:
            self.style.node_hover = None
        elif arg is True:
            ordered_features = ["idx", "dist", "support", "height"]
            lfeatures = list(set(self.ttree.features) - set(ordered_features))       
            ordered_features += lfeatures

            # build list of hoverstrings in order of idxs
            self.style.node_hover = [" "] * self.ttree.nnodes
            for idx in self.ttree.idx_dict:
                feats = []
                node = self.ttree.idx_dict[idx]
                for feature in ordered_features:
                    val = getattr(node, feature)
                    if isinstance(val, float):
                        feats.append("{}: {:.4f}".format(feature, val))
                    else:
                        feats.append("{}: {}".format(feature, val))
                self.style.node_hover[idx] = "\n".join(feats)
            # already in levelorder from idx_dict iter
            self.style.node_hover = self.style.node_hover

        # project to size
        self.style.node_hover = toyplot.broadcast.pyobject(
            self.style.node_hover, self.nnodes)

        # special: hide nodes with labels that are (TODO: nan) for now ("")
        # node_labels is already in levelorder.
        self.style.node_hover[self.style.node_labels == ""] = None


    def _assign_node_colors(self):
        """
        Sets .node_colors as an array of length nedges. Default is to 
        leave the list as None so that the default style fill all.

        node_colors=None         | [None, None, None, ...]
        node_colors='red'        | [None, None, ...] & sets node_style['fill']
        node_colors=ITERABLE     | ['red', 'blue', 'green']
        """
        arg = self.style.node_colors

        # set to empty 
        if arg is None:
            self.style.node_colors = toyplot.broadcast.pyobject(arg, self.nnodes)

        # fill array with whatever was provided
        else:
            self.style.node_colors = toyplot.color.broadcast(arg, self.nnodes)[::-1]

            # if all the same then reset to None
            if len(set([str(i) for i in self.style.node_colors])) == 1:

                # save the fixed color and set to None
                color = self.style.node_colors[0]
                self.style.node_colors = toyplot.broadcast.pyobject(None, self.nnodes)

                # set edge_style stroke and stroke-opacity 
                sub = split_rgba_style({'fill': color})
                self.style.node_style['fill'] = sub["fill"]
                self.style.node_style['fill-opacity'] = sub["fill-opacity"]


    def _assign_edge_colors(self):
        """
        Sets .edge_colors as an array of length nedges. Default is to 
        leave the list as None so that the default style stroke fills all.

        edge_colors=None         | [None, None, None, ...]
        edge_colors='red'        | [None, None, ...] & sets edge_style['fill']
        edge_colors=ITERABLE     | ['red', 'blue', 'green']
        """
        arg = self.style.edge_colors

        # set to empty 
        if arg is None:
            self.style.edge_colors = toyplot.broadcast.pyobject(arg, self.nedges)

        # fill array with whatever was provided
        else:
            self.style.edge_colors = toyplot.color.broadcast(arg, self.nedges)[::-1]

            # if all the same then reset to None
            if len(set([str(i) for i in self.style.edge_colors])) == 1:

                # save the fixed color and set to None
                color = self.style.edge_colors[0]
                self.style.edge_colors = toyplot.broadcast.pyobject(None, self.nedges)

                # set edge_style stroke and stroke-opacity 
                sub = split_rgba_style({'stroke': color})
                self.style.edge_style['stroke'] = sub["stroke"]

                # only set sub opacity if it is not already at 1, otherwise
                # keep the edgestyledict opacity b/c user probably set it.
                if self.style.edge_style["stroke-opacity"] == 1:
                    self.style.edge_style['stroke-opacity'] = sub["stroke-opacity"]


    def _assign_edge_widths(self):
        """
        Sets .edge_widths as an array of length nedges. Default is to 
        leave the list as None so that the default style stroke-width sets
        all values. There is special option for entry "Ne" which will auto
        scale edge withs by the range of values on TreeNodes.Ne

        edge_widths=None         | [None, None, None, ...]
        edge_widths=2            | [None, None, ...] & edge_style[stroke-width]
        edge_widths=ITERABLE     | [2, 3, 3, 2, ...]
        edge_widths='Ne'         | [2, 2.5, 5, 2, ...]
        """
        arg = self.style.edge_widths

        # set to empty 
        if arg is None:
            self.style.edge_widths = toyplot.broadcast.pyobject(arg, self.nedges)

        # fill array with whatever was provided
        else:
            if str(arg) == "Ne":
                try:
                    arg = self.ttree.get_edge_values("Ne", normalize=True)
                except Exception:
                    arg = np.repeat(2, self.nedges)
            self.style.edge_widths = toyplot.broadcast.pyobject(arg, self.nedges)[::-1]

            # if all the same then reset to None
            if len(set([str(i) for i in self.style.edge_widths])) == 1:

                # save the fixed color and set to None
                width = self.style.edge_widths[0]
                self.style.edge_widths = toyplot.broadcast.pyobject(None, self.nedges)

                # set edge_style stroke and stroke-opacity 
                self.style.edge_style['stroke-width'] = width



    def _assign_admixture_idxs(self):
        """
        Check the input admixture arg list and expand/check so 
        that final format is: 

        admixture_edges = [
            (src_idx, dest_idx, (src_time, dest_time), styledict, label)
        ]
        """
        arg = self.style.admixture_edges

        # bail if empty
        if arg is None:
            return 

        # if tuple then expand into a list
        if isinstance(arg, tuple):
            arg = [arg]

        colors = itertools.cycle(COLORS1)
        next(colors)
        admix_tuples = []
        for atup in arg:

            # required src, dest
            if isinstance(atup[0], (str, list, tuple)):
                nas = NodeAssist(self.ttree, atup[0], None, None)
                node = nas.get_mrca()
                if not node.is_root():
                    src = node.idx
                else:
                    nas.match_reciprocal()
                    src = nas.get_mrca().idx
            else:
                src = int(atup[0])

            if isinstance(atup[1], (str, list, tuple)):
                nas = NodeAssist(self.ttree, atup[1], None, None)
                node = nas.get_mrca()
                if not node.is_root():
                    dest = node.idx
                else:
                    nas.match_reciprocal()
                    dest = nas.get_mrca().idx
            else:
                dest = int(atup[1])

            # optional: proportion on edges
            if len(atup) > 2:
                if isinstance(atup[2], (int, float)):
                    prop = float(atup[2])
                else:
                    prop = (float(atup[2][0]), float(atup[2][1]))
            else:
                prop = 0.5

            # optional: style dictionary
            if len(atup) > 3:
                style = dict(atup[3])
            else:
                style = {}

            # optional label on midpoint
            if len(atup) > 4:
                label = str(atup[4])
            else:
                label = None

            # color setting and correction
            if "stroke" not in style:
                style['stroke'] = next(colors)

            if "stroke-opacity" not in style:
                style['stroke-opacity'] = '0.7'

            # colorfix to edge styles
            colorfix = split_rgba_style(style.copy())
            style['stroke'] = colorfix['stroke']

            # check styledict colors, etc
            admix_tuples.append((src, dest, prop, style, label))


        self.style.admixture_edges = admix_tuples



# #!/usr/bin/env python

# """DEPRECATED FOR add_scale_bar and ...

# """

# import numpy as np
# import toyplot
# from toytree import ToyTree
# from toytree.core import Cartesian
# from toytree.core.apis import add_subpackage_method, AnnotationAPI
# from toytree.annotate.src.annotation_mark import (
#     get_last_toytree_mark_from_cartesian,
#     assert_tree_matches_mark,
# )


# def set_axes_ticks_style(
#     tree: ToyTree,
#     axes: Cartesian,
#     only_inside: bool = True,
#     **kwargs,
# ) -> Cartesian:
#     """Return a Cartesian axes modified for custom tick marks.

#     This gets tick locations first using toyplot.locator.Extended and
#     then sets labels on them using toyplot.locator.Explicit, because
#     we need time scale bar to be non-negative when axes are rotated
#     for trees facing different directions.

#     Note
#     -----
#     Some work is done internally to try to nicely handle floating point
#     precision.

#     Parameters
#     ----------
#     ...
#     style: TreeStyle
#         A TreeStyle object with options for styling axes.
#     only_inside: bool
#         Option used by toyplot.locator.Extended to automatically find
#         tick marks given the data range.
#     """
#     mark = get_last_toytree_mark_from_cartesian(axes)
#     assert_tree_matches_mark(tree, mark)

#     # the axes is either new or passed as an arg, and the scale_bar
#     # arg is True or a (float, int), so we need to style the ticks.
#     if mark.layout in ("r", "l"):
#         nticks = max((4, np.floor(mark.width / 75).astype(int)))
#         axes.y.show = False
#         axes.x.show = True
#         axes.x.ticks.show = True
#     elif mark.layout in ("u", "d"):
#         nticks = max((4, np.floor(mark.height / 75).astype(int)))
#         axes.x.show = False
#         axes.y.show = True
#         axes.y.ticks.show = True
#     # e.g., unrooted layout with axes shown (e.g., ts='p')
#     else:
#         # nticks = max((4, np.floor(style.height / 75).astype(int)))
#         nticks = 5
#         axes.x.show = False
#         axes.y.show = False

#     # get tick locator
#     lct = toyplot.locator.Extended(count=nticks, only_inside=only_inside)

#     # get root tree height
#     if mark.layout in ("r", "u"):
#         locs = lct.ticks(-tree_height, -0)[0]
#     else:
#         locs = lct.ticks(0, tree_height)[0]

#     # apply unit scaling
#     if mark.scale_bar is False:
#         labels = abs(locs.copy())
#     elif isinstance(mark.scale_bar, (int, float)):
#         labels = abs(locs / mark.scale_bar)
#     else:
#         labels = abs(locs.copy())
#     labels = [np.format_float_positional(i, precision=6, trim="-") for i in labels]

#     # set the ticks locator
#     if mark.layout in ("r", "l"):
#         axes.x.ticks.locator = toyplot.locator.Explicit(
#             locations=locs + mark.xbaseline,
#             labels=labels,
#         )
#     elif mark.layout in ("u", "d"):
#         axes.y.ticks.locator = toyplot.locator.Explicit(
#             locations=locs + mark.ybaseline,
#             labels=labels,
#         )
#     # print(locs, labels)
#     return axes



#!/usr/bin/env python

"""..."""

from typing import Union
from pathlib import Path
import re
import toytree


def parse_network(net: Union[str, Path], disconnect=True):
    """Parse a network file to extract a major topology and admix dict.

    This leaves the hybrid nodes in the tree and labels each with 
    .name="H{int}" and .gamma={float}.
    """
    # if net is a file then read the first line
    if ";" not in net:
        if not net.exists():
            raise IOError("input type is not a file or newick (no ; ending).")
        with open(net, 'rt', encoding="utf-8") as infile:
            net = infile.readline()
    else:
        # trim off loglik and anything after it (TODO: keep loglik)
        if ";" in net:
            net = net.split(";")[0] + ';'

    # sub :xxx:: to be ::: b/c I don't care about admix edge bls FOR NOW.
    net = re.sub(r":\d.\w*::", ":::", net)
    print(net)

    # change H nodes to be internal node labels rather than tip nodes.
    while ",#" in net:
        pre, post = net.split(",#", 1)
        npre, npost = post.split(")", 1)
        newpre = npre.split(":")[0] + "-" + npre.split(":")[-1]
        net = pre + ")#" + newpre + npost
    net = net.replace(":::", "-")
    print(net)

    # parse cleaned newick and set empty gamma on all nodes
    net = toytree.tree(net)#, tree_format=1)
    print(net.get_node_data())

    # store admix data
    admix = {}

    # if not rooted choose any non-H root
    if not net.is_rooted():
        net = net.root(
            [i for i in net.get_tip_labels() if not i.startswith("#H")][0]
        )

    # Traverse tree to find hybrid nodes. If a hybrid node is labeled as a 
    # distinct branch in the tree then it is dropped from the tree and 
    for node in net.traverse("postorder"):

        # find hybrid nodes as internal nchild=1, or external with H in name
        if (len(node.children) == 1) or node.name.startswith("#H"):

            # assign name and gamma to hybrid nodes
            print(node.name)
            aname, aprop = node.name.split("-")
            aname = aname.lstrip("#")
            node.name = aname

            # assign hybrid to closest nodes up and down from edge
            # node.children[0].hybrid = int(aname[1:])
            # node.gamma = round(float(aprop), 3)
            # node.up.hybrid = int(aname[1:])

            # if root is a hybrid edge (ugh)
            if node.up is None:
                small, big = sorted(node.children, key=len)
                root = toytree.Node(name='root')
                node.children = [small]
                small.up = node
                node.up = root
                big.up = root
                root.children = [node, big]
                net.treenode = root

            # disconnect node by connecting children to parent
            if disconnect:

                # if tip is a hybrid
                if not node.children:
                    # get sister node
                    sister = [i for i in node.up.children if i != node][0]

                    # connect sister to gparent
                    sister.up = node.up.up
                    node.up.up._remove_child(node.up)
                    node.up.up._add_child(sister)

                # if hybrid is internal
                else:
                    node.up._remove_child(node)
                    for child in node.children:
                        child._up = node.up
                        node.up._add_child(child)

            # store admix data by descendants but remove hybrid tips
            desc = node.get_leaf_names()
            if aname in desc:
                desc = [i for i in node.up.get_leaf_names() if i != aname]
            desc = [i for i in desc if not i.startswith("#H")]

            # put this node into admix
            if aname not in admix:
                admix[aname] = (desc, aprop)

            # matching edge in admix, no arrange into correct order by minor
            else:
                # this is the minor edge
                if aprop < admix[aname][1]:
                    admix[aname] = (
                        admix[aname][0], 
                        desc, 
                        0.5, 
                        {}, 
                        str(round(float(aprop), 3)),
                    )

                # this is the major edge
                else:
                    admix[aname] = (
                        desc, 
                        admix[aname][0], 
                        0.5, 
                        {}, 
                        str(round(float(admix[aname][1]), 3)),
                    )

    # update coords needed if node disconnection is turned back on.
    # net._coords.update()
    net._update()
    net = net.mod.ladderize()
    return net, admix



#!/usr/bin/env python

"""Annotation functions for adding to toytree drawings.
"""


class Annotator(object):
    """
    Add annotations as a new mark on top of an existing toytree mark.
    """
    def __init__(self, tree, axes, mark):
        self.tree = tree
        self.axes = axes
        self.mark = mark

    def draw_clade_box(
        self, 
        names=None, 
        regex=None, 
        wildcard=None, 
        yspace=None, 
        xspace=None, 
        **kwargs):
        """
        Draw a rectangle around a clade on a toytree.

        Parameters:
        -----------
        names, regex, wildcard:
            Choose one of these three methods to select one or more tipnames. 
            The clade composing all descendants of their common ancestor will 
            be highlighted.

        yspace (float or None):
            The extent to which boxes extend above and below the root and tip
            nodes. If None then this is automatically generated.

        xspace (float or None):
            The extent to which the clade box extends to the sides 
            (out of the clade towards other tips.) If None default uses 0.5.

        kwargs:
            Additional styling options are supported: color, opacity, etc.

        Returns:
        ------------
        Toyplot.mark.Range
        """

        # get the common ancestor
        nidx = self.tree.get_mrca_idx_from_tip_labels(
            names=names, regex=regex, wildcard=wildcard)

        # get tips descended from mrca
        tips = self.tree[nidx].get_leaves()
        tidxs = [i.idx for i in tips]

        # extent to which box bounds extend outside of the exact clade size.
        if not yspace:
            yspace = self.tree.treenode.height / 15.
        if not xspace:
            xspace = 0.45

        # left and right positions
        if self.mark.layout == 'r':
            xmin = self.mark.ntable[nidx, 0] - yspace
            xmax = max(self.mark.ntable[tidxs, 0]) + yspace
            ymin = min(self.mark.ntable[tidxs, 1]) - xspace
            ymax = max(self.mark.ntable[tidxs, 1]) + xspace   

        if self.mark.layout == 'l':
            xmin = self.mark.ntable[nidx, 0] + yspace
            xmax = max(self.mark.ntable[tidxs, 0]) - yspace
            ymin = max(self.mark.ntable[tidxs, 1]) + xspace
            ymax = min(self.mark.ntable[tidxs, 1]) - xspace   

        elif self.mark.layout == 'd':
            ymax = self.mark.ntable[nidx, 1] + yspace
            ymin = min(self.mark.ntable[tidxs, 1]) - yspace
            xmin = min(self.mark.ntable[tidxs, 0]) - xspace
            xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

        elif self.mark.layout == 'u':
            ymin = self.mark.ntable[nidx, 1] - yspace
            ymax = min(self.mark.ntable[tidxs, 1]) + yspace
            xmin = min(self.mark.ntable[tidxs, 0]) - xspace
            xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

        # draw the rectangle
        newmark = self.axes.rectangle(xmin, xmax, ymin, ymax, **kwargs)

        # put tree at the top of the scenegraph
        self.axes._scenegraph.remove_edge(self.axes, 'render', self.mark)
        self.axes._scenegraph.add_edge(self.axes, 'render', self.mark)

        return newmark



    # def draw_tip_box(
    #     self, 
    #     names=None, 
    #     regex=None, 
    #     wildcard=None, 
    #     yspace=None, 
    #     xspace=None, 
    #     **kwargs):
    #     """
    #     Draw a rectangle around the tips of a clade on a toytree.

    #     Parameters:
    #     -----------
    #     names, regex, wildcard:
    #         Choose one of these three methods to select one or more tipnames. 
    #         The clade composing all descendants of their common ancestor will 
    #         be highlighted.

    #     yspace (float or None):
    #         The extent to which boxes extend above and below the root and tip
    #         nodes. If None then this is automatically generated.

    #     xspace (float or None):
    #         The extent to which the clade box extends to the sides 
    #         (out of the clade towards other tips.) If None default uses 0.5.

    #     kwargs:
    #         Additional styling options are supported: color, opacity, etc.

    #     Returns:
    #     ------------
    #     Toyplot.mark.Range
    #     """

    #     # get the common ancestor
    #     nidx = self.tree.get_mrca_idx_from_tip_labels(
    #         names=names, regex=regex, wildcard=wildcard)

    #     # get tips descended from mrca
    #     tips = self.tree[nidx].get_leaves()
    #     tidxs = [i.idx for i in tips]

    #     # get nudge size from dists in the tree or user supplied
    #     if not yspace:
    #         yspace = self.tree.get_node_values("dist", 1, 1).mean() / 4.
    #     if not xspace:
    #         xspace = 0.5

    #     # distance in PIXELS to the tip labels
    #     tipx = toyplot.units.convert(
    #         mark.tip_labels_style["-toyplot-anchor-shift"], 'px')

    #     # left and right positions
    #     if self.mark.layout == 'r':           

    #         # get unit conversion
    #         tipstart = tipx / (axes.project('x', 1) - axes.project('x', 0))
    #         xmin = self.mark.ntable[nidx, 0] - yspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) + yspace
    #         ymin = min(self.mark.ntable[tidxs, 1]) - xspace
    #         ymax = max(self.mark.ntable[tidxs, 1]) + xspace   

    #     if self.mark.layout == 'l':
    #         xmin = self.mark.ntable[nidx, 0] + yspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) - yspace
    #         ymin = max(self.mark.ntable[tidxs, 1]) + xspace
    #         ymax = min(self.mark.ntable[tidxs, 1]) - xspace   

    #     elif self.mark.layout == 'd':
    #         ymax = self.mark.ntable[nidx, 1] + yspace
    #         ymin = min(self.mark.ntable[tidxs, 1]) - yspace
    #         xmin = min(self.mark.ntable[tidxs, 0]) - xspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

    #     elif self.mark.layout == 'u':
    #         ymin = self.mark.ntable[nidx, 1] - yspace
    #         ymax = min(self.mark.ntable[tidxs, 1]) + yspace
    #         xmin = min(self.mark.ntable[tidxs, 0]) - xspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

    #     # draw the rectangle
    #     mark = self.axes.rectangle(xmin, xmax, ymin, ymax, **kwargs)
    #     return mark



    # def generate_rectangle(self, firstname=None, lastname=None, axes=None, color="green", opacity=.25):
    #     """
    #     Returns an updated axes with a generated rectangle based on input labels provided
    #     """
    #     index_of_first = self.get_mrca_idx_from_tip_labels(names=firstname)
    #     index_of_last =  self.get_mrca_idx_from_tip_labels(names=lastname)
    #     x_vals = (x[0] for x in self.get_node_coordinates())

    #     axes.rectangle(
    #         min(self.get_tip_coordinates()[index_of_first][0],  self.get_tip_coordinates()[index_of_last][0]),
    #         max(x_vals),
    #         self.get_tip_coordinates()[index_of_first][1],
    #         self.get_tip_coordinates()[index_of_last][1],
    #         opacity=opacity,
    #         color=color,
    #     )
    #     return axes




#!/usr/bin/env python

"""Wrapper that allows wrapping a method to a class.

"""

from functools import wraps


def add_cartesian_method(cls):
    """Clever approach to copy docstring and signature from func to API.

    Reference
    ---------
    https://mgarod.medium.com/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
    """
    def decorator(func):

        # creates a wrapper for a Cartesian.{function} that copies
        # its docstring and arg signatures but sets axes as the
        # first argument of the function.
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        # sets: TreeModAPI.{function} = wrapper
        # TODO: remove first arg from signature and docs.
        setattr(cls, func.__name__, wrapper)

        # Note we are not binding func, but wrapper which accepts self
        # but does exactly the same as func.
        # returning func means func can still be used normally
        return func
    return decorator


if __name__ == "__main__":

    from toytree.utils.src.toyplotfix.cartesian import Cartesian

    @add_cartesian_method(Cartesian)
    def test(axes: Cartesian, arg2: str) -> float:
        """This is a test."""
        return 32.0

    help(test)
    help(Cartesian.test)



#!/usr/bin/env python


import toytree


def logo():
    tree = toytree.rtree.unittree(5, seed=300)
    c, a, m = tree.draw(
        width=150, height=180,
        layout='d',
        edge_type='c', edge_widths=3, tip_labels=False,
    )
    toytree.annotate.add_tip_markers(tree, a, size=[4, 5, 6, 5, 4], style={"stroke": None}, yshift=8, color=toytree.color.COLORS1[0])
    toytree.annotate.add_tip_markers(tree, a, size=[6, 5, 4, 3, 4], style={"stroke": None}, yshift=16, color=toytree.color.COLORS1[1])
    toytree.annotate.add_tip_markers(tree, a, size=[4, 3, 4, 5, 6], style={"stroke": None}, yshift=24, color=toytree.color.COLORS1[2])
    toytree.annotate.add_tip_markers(tree, a, size=[5, 6, 7, 6, 5], style={"stroke": None}, yshift=32, color=toytree.color.COLORS1[3])
    toytree.annotate.add_tip_markers(tree, a, size=[6, 5, 4, 5, 6], style={"stroke": None}, yshift=40, color=toytree.color.COLORS1[4])
    return c, a, m


if __name__ == "__main__":
    logo()



#!/usr/bin/env python

"""
PCM: phylogenetic comparative methods tools
"""

import os
import time
import numpy as np
import toytree



class PCM:
    """
    Phylogenetic comparative methods implemented on toytrees.
    """
    def __init__(self, tree):
        self.tree = tree


    def independent_contrasts(self, feature):
        ntree = self.tree.copy()
        resdict = PIC(ntree, feature)
        ntree = ntree.set_node_values(
            feature="{}-contrast",
            values={i.name: j[2] for (i, j) in resdict.items()}
        )
        ntree = ntree.set_node_values(
            feature="{}-contrast-var",
            values={i.name: j[3] for (i, j) in resdict.items()}
        )        
        return ntree


    def ancestral_state_reconstruction(self, feature):
        """
        Infer ancestral states on ancestral nodes for continuous traits
        under a brownian motion model of evolution.

        Modified from IVY interactive (https://github.com/rhr/ivy/)

        Returns a toytree with feature applied to each node.
        """
        ntree = self.tree.copy()
        resdict = PIC(ntree, feature)
        ntree = ntree.set_node_values(
            feature, 
            values={i.name: j[0] for (i, j) in resdict.items()}
        )
        return ntree


    def tree_to_VCV(self):
        return VCV(self.tree)




def VCV(tree):
    """
    Return the variance co-variance metrix representing the tree topology.
    """
    vcv_ = np.zeros((tree.ntips,tree.ntips))
    labs = tree.get_tip_labels()
    for lab1 in range(tree.ntips):
        for lab2 in range(tree.ntips):
            mrca_idx = tree.get_mrca_idx_from_tip_labels([labs[lab1],labs[lab2]])
            mrca_height = tree.treenode.search_nodes(idx=mrca_idx)[0].height
            vcv_[lab1, lab2] = tree.treenode.height - mrca_height
    return(vcv_)



def PIC(tree, feature):
    """
    Infer ancestral states and calculate phylogenetic independent
    contrasts at nodes for a selected feature (trait). 

    Modified from IVY interactive (https://github.com/rhr/ivy/)

    Parameters
    ----------
    feature: (str)
        The name of a feature of the tree that has been mapped to all 
        tip nodes of the tree. 

    Returns
    -------
    toytree (Toytree.ToyTree)
        A modified copy of the input tree is returned with the mean ancestral
        value for the selected feature inferred for all nodes of the tree. 
    """
    # get current node features at the tips
    fdict = tree.get_feature_dict(key_attr="name", values_attr=feature)
    data = {i: j for (i, j) in fdict.items() if i in tree.get_tip_labels()}

    # apply dynamic function from ivy to return dict results
    results = dynamicPIC(tree.treenode, data, results={})

    # return dictionary mapping nodes to (mean, var, contrast, cvar)
    return results



def dynamicPIC(node, data, results):
    """
    Phylogenetic independent contrasts. Recursively calculate 
    independent contrasts of a bifurcating node given a dictionary
    of trait values.

    Modified from IVY interactive (https://github.com/rhr/ivy/)

    Args:
        node (Node): A node object
        data (dict): Mapping of leaf names to character values
    Returns:
        dict: Mapping of internal nodes to tuples containing ancestral
              state, its variance (error), the contrast, and the
              contrasts's variance.
    TODO: modify to accommodate polytomies.
    """    
    X = []
    v = []

    # recursively does children until X and v are full
    for child in node.children:

        # child has children, do those first
        if child.children:

            # update results dict with children values
            dynamicPIC(child, data, results)
            child_results = results[child]

            # store childrens values
            X.append(child_results[0])
            v.append(child_results[1])

        # no child of child, so just do child
        else:
            X.append(data[child.name])
            v.append(child.dist)

    # Xi - Xj is the contrast value
    Xi, Xj = X  

    # vi + vj is the contrast variance
    vi, vj = v

    # Xk is the reconstructed state at the node
    Xk = ((1.0 / vi) * Xi + (1 / vj) * Xj) / (1.0 / vi + 1.0 / vj)

    # vk is the variance
    vk = node.dist + (vi * vj) / (vi + vj)

    # store in dictionary and 
    results[node] = (Xk, vk, Xi - Xj, vi + vj)
    return results



def calculate_ES(tree):
    "Return DataFrame with equal splits measure sensu Redding and Mooers 2006"
    # dataframe for storing results
    df = pd.DataFrame(columns=["DR"], index=tree.get_tip_labels())

    # traverse up to root from each tip
    for idx in range(tree.ntips):
        node = tree.idx_dict[idx]
        DR = 0
        j = 1
        while node.up:
            DR += node.up.dist / (2 ** j)
            node = node.up
            j += 1
        df.iloc[idx, 0] = DR
    return df



def calculate_DR(tree):
    "Returns a dataframe with tip-level diversification rates sensu Jetz 2012"
    # ensure tree is a tree
    tree = toytree.tree(tree)
    return 1 / calculate_ES(tree)



def calculate_tip_level_diversification(trees, njobs=1):
    """
    Returns a dataframe with tip-level diversification rates calculated
    across a set of trees. Enter a multitree object or, for very large trees,
    you can enter a generator of newick strings (see example).

    Parameters:
    -----------
    trees (Toytree, Multitree, or newick file):
        One or more ultrametric trees in newick format. For ultra large
        datasets use a file as input.

    njobs (int):
        Distribute N jobs in parallel usign ProcessPoolExecutor

    Returns:
    --------
    df (pandas.DataFrame):
        The raw DF based on branch lengths of the tree.

    Examples:
    ---------
    # using one or more trees
    df = calculate_tip_level_diversification(trees)

    # loading super large trees
    with open(hugetreefile, 'r') as treegenerator:
        df = calculate_tip_level_diversification(treegenerator, njobs=20)
    """

    # not yet adding pandas as a global dependency
    import pandas as pd

    # load data and metadata from newick, toytree, or multitree
    if isinstance(trees, str) and os.path.exists(trees):
        with open(trees) as tree_generator:
            tre = toytree.tree(next(tree_generator))
            ntips = tre.ntips
            ntrees = sum(1 for i in tree_generator) + 1
            tiporder = tre.get_tip_labels()
        itertree = open(trees, 'r')
    elif isinstance(trees, toytree.Toytree.ToyTree):
        itertree = iter([trees])
        ntrees = len(trees)
        ntips = trees.ntips
        tiporder = trees.get_tip_labels()
        topen = None
    elif isinstance(trees, toytree.Multitree.MultiTree):
        itertree = iter(i for i in trees)
        ntrees = len(trees)
        ntips = trees.treelist[0].ntips
        tiporder = trees.treelist[0].get_tip_labels()
        topen = None
    else:
    	raise IOError("problem with input: {}".format(trees))

    # array to store results 
    tarr = np.zeros((ntips, ntrees))

    # run non-parallel calculations
    if njobs == 1:
        for tidx, tree in enumerate(itertree):
            df = calculate_DR(tree)
            tarr[:, tidx] = df.loc[tiporder, "DR"]

    # or, distribute jobs in parallel (py3 only)
    else:
        from concurrent.futures import ProcessPoolExecutor 
        pool = ProcessPoolExecutor(njobs)
        treegen = iter(trees)
        rasyncs = {}

        # submit as many jobs as there are cores
        for job in range(njobs):
            tree = next(itertree)
            rasyncs[job] = pool.submit(
                calculate_DR,
                (tree),
            )

        # as each job finished submit a new one until none are left.
        # this allows avoiding loading all trees simultaneously.
        tidx = 0
        while 1:
            # get finished jobs
            finished = [i for i in rasyncs if rasyncs[i].done()]
    
            # store results and append new job to the engine
            for job in finished:

                # store result
                result = rasyncs[job].result()
                tarr[:, tidx] = result.loc[tiporder, "DR"]
                tidx += 1
                del rasyncs[job]

                # append new job unless no jobs left
                try:
                    tree = next(itertree)
                    rasyncs[job] = pool.submit(
                        calculate_DR,
                        (tree),
                    )
                except StopIteration:
                    pass

            # wait for next check
            time.sleep(0.5)           

            # wait until all jobs finish
            if not rasyncs:
                break
        pool.shutdown()

    # close the file handle, if exists
    if hasattr(itertree, "close"):
        itertree.close()

    # calculate summary of DR across the distribution of trees.
    arr = pd.DataFrame(tarr, index=tiporder)
    df = pd.DataFrame({
        'mean': arr.mean(1),
        # 'harmMean': hmean(arr, axis=1),
        'median': arr.median(1),
        'std': arr.std(1),
        '2.5%': np.percentile(arr, 0.025, axis=1),
        '97.5%': np.percentile(arr, 0.975, axis=1),
        'min': arr.min(1),
        'max': arr.max(1),
    })
    return df




# single test
if __name__ == "__main__":

    import toyplot
    import toytree

    colormap = toyplot.color.brewer.map("BlueRed", reverse=True)
    colormap

    tree = toytree.rtree.imbtree(5, 1e6)
    tree = tree.set_node_values(
        "g", 
        values={i: 5 for i in (2, 3, 4)},
        default=1,
    )
    tree.draw(
        ts='p', 
        node_labels=tree.get_node_values("g", 1, 1),
        node_colors=[
            colormap.colors(i, 0, 5) for i in tree.get_node_values('g', 1, 1)]
        )

    # apply reconstruction
    ntree = PIC(tree, "g")

    # new values are stored as -mean, -var, -contrasts, ...
    evals = ntree.get_edge_values("g-mean")

    # draw new tree
    ntree.draw(
        ts='p', 
        node_labels=ntree.get_node_values("g-mean", 1, 1),
        node_colors=[
            colormap.colors(i, 0, 5) for i in 
            ntree.get_node_values('g-mean', 1, 1)]
    )



#!/usr/bin/env python

from setuptools import setup, find_packages
import re
from pathlib import Path

DIR = Path(__file__).parent
LONG_DESC = (DIR / "README.md").read_text()
INIT_DATA = (DIR / "toytree/__init__.py").read_text()
VERSION = re.search(
    r"^__version__ = ['\"]([^'\"]*)['\"]",
    INIT_DATA,
    re.M,
).group(1)

# get version from __init__.py
#INITFILE = "toytree/__init__.py"
#CUR_VERSION = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
#                    open(INITFILE, "r").read(),
#                    re.M).group(1)

# run setup
setup(
    name="toytree",
    version=VERSION,
    url="https://github.com/eaton-lab/toytree",
    author="Deren Eaton",
    author_email="de2356@columbia.edu",
    description="minimalist tree plotting using toyplot",
    long_description=LONG_DESC,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        "toyplot",
        "numpy",
        "requests",
        "future",
    ],
    entry_points={},
    license='GPL',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
)



#!/usr/bin/env python

"""
Classes for Drawings from MultiTrees.
"""
from decimal import Decimal
import numpy as np
import toyplot
from .TreeStyle import TreeStyle

from toyplot.mark import Mark
import xml.etree.ElementTree as xml

# Register multipledispatch to share with toyplot.html
import functools
from multipledispatch import dispatch
dispatch = functools.partial(dispatch, namespace=toyplot.html._namespace)




class CloudTreeMark(Mark):
    """
    
    """


class RenderCloudTree:
    """

    """


@dispatch(toyplot.coordinates.Cartesian, CloudTreeMark, toyplot.html.RenderContext)
def _render(axes, mark, context):
    RenderCloudTree(axes, mark, context)




class TreeGridMark(Mark):
    """
    Multiple toytrees on the same cartesian coordinates axes but spread 
    out non-overlapping on a grid using layout plans. Applies shared tree
    styles to a top-level treegrid to reduce repetition in CSS.
    """
    def __init__(self, trees):

        # inherit type
        Mark.__init__(self)

        # store all needed from subtrees.
        self._coordinate_axes = ['x', 'y']

        # the tree with their styles which will make tree marks
        # from which contex will be extracted 
        self.trees = trees

        # grid layout arguments




class RenderTreeGrid:
    """
    Organized class to call in _render to create a single Mark with 
    multiple trees arranged into a grid using xbaseline, ybaseline, 
    and either scaling trees to fit nicely or using original scales
    (e.g., different root heights) to show differences. 

    Main usage:
        1. comparing topologies (down or right facing; scaled)
        2. comparing coalescence (down facing; not scaled).


    """
    def __init__(self, axes, mark, context):

        # inputs: mark has .trees and its own top-level .style.
        # axes and context/canvas are singular.
        self.mark = mark
        self.axes = axes
        self.context = context

        # to be constructed
        self.mark_xml = None

        # construction funcs
        self.project_grid()
        self.build_dom()


    def build_dom(self):

        # get shared styles


        # build top level dom
        self.mark_treegrid()

        # lay down trees
        self.mark_toytrees()



    def project_grid(self):
        """
        Figure out the grid layout based on the size of each 
        tree, the user-provided width/height, and shared axes args.
        """

        # option 1: no shared axes. Trees are scaled(?)


        # option 2: shared axes. Spacing of tree



    def mark_treegrid(self):
        """
        Creates the top-level Toytree-Grid mark.
        """
        self.mark_xml = xml.SubElement(
            self.context.parent, "g", 
            id=self.context.get_id(self.mark),
            attrib={"class": "toytree-mark-Grid"},
        )


    def mark_toytrees(self):
        """
        Creates toytree marks and appends them in the Grid Mark
        """
        for tre in self.trees:
            # render the edge group
            self.tree_xml = xml.SubElement(
                self.mark_xml, "g", 
                attrib={"class": "toytree-mark-Toytree"}, 
                style=style_to_string(self.mark.edge_style)
            )





@dispatch(toyplot.coordinates.Cartesian, TreeGridMark, toyplot.html.RenderContext)
def _render(axes, mark, context):
    RenderTreeGrid(axes, mark, context)






# class OldTreeGrid(object):
#     """
#     Easily create Toyplot gridded canvases for plotting multiple trees.
#     """
#     def __init__(self, treelist):

#         # plot objects are init on update()
#         self.canvas = None
#         self.treelist = treelist
#         self.treeslice = []

#         # to be filled
#         self.nrows = None
#         self.ncols = None


#     def update(self, axes, nrows, ncols, start, shared_axis, **kwargs):

#         # store plot dims and assert that they fit well enough
#         self.axes = axes
#         self.nrows = nrows
#         self.ncols = ncols
#         self.treeslice = self.treelist[start:start + self.nrows * self.ncols]

#         # TODO: mess with padding and margins...
#         if not self.axes:
#             wdef = min(800, self.ncols * 175)
#             hdef = min(800, self.nrows * 250)
#             self.canvas = toyplot.Canvas(
#                 width=(kwargs.get('width') if kwargs.get('width') else wdef), 
#                 height=(kwargs.get('height') if kwargs.get('height') else hdef),
#             ) 

#         if (not shared_axis) and (not self.axes):
#             # get max treeheight
#             for tidx, tree in enumerate(self.treeslice):

#                 # create grid with a reasonable margin between trees
#                 #if not axes:
#                 axes = self.canvas.cartesian(
#                     grid=(self.nrows, self.ncols, tidx),
#                     margin=(30, 30, 30, 30),
#                     padding=35,
#                 )
#                 axes.show = False

#                 # update tree style with any new arguments
#                 tree.draw(axes=axes)


#         # shared X axis
#         else:    
#             # only one axis allowed?... x=1 or y=1
#             if not self.axes:
#                 axes = self.canvas.cartesian(padding=25)
#             xbaseline = 0
#             maxheight = 0
#             layout = ("d" if not kwargs.get("layout") else kwargs.get("layout"))

#             # COALESCENT TIME COMPARISON TYPE PLOT
#             if layout == "d":
#                 for tidx, tree in enumerate(self.treeslice):
#                     tree.draw(axes=axes, xbaseline=xbaseline, layout='d')
#                     xbaseline += tree.ntips + 1
#                     maxheight = max(maxheight, tree.treenode.height)

#                 if not self.axes:
#                     nticks = max((3, np.floor(hdef / 100).astype(int)))
#                     axes.x.show = False
#                     axes.y.show = True
#                     axes.y.ticks.show = True            

#                     # generate locations
#                     locs = np.linspace(0, maxheight, nticks)

#                     # auto-formatter for axes ticks labels
#                     zer = abs(min(0, Decimal(locs[1]).adjusted()))
#                     fmt = "{:." + str(zer) + "f}"
#                     axes.y.ticks.locator = toyplot.locator.Explicit(
#                         locations=locs,
#                         labels=[fmt.format(i) for i in np.abs(locs)],
#                         )

#             # USUALLY TOPOLOGY COMPARISON PLOT
#             else:
#                 for tidx, tree in enumerate(self.treeslice):
#                     tree = tree.mod.node_scale_root_height(1.0)
#                     tree.draw(axes=axes, xbaseline=xbaseline, layout='r')

#                     if kwargs.get('xbaseline'):
#                         xbaseline += kwargs.get('xbaseline')
#                     else:
#                         if tree.style.use_edge_lengths is False:
#                             shift = tree.treenode.get_farthest_leaf(
#                                 topology_only=True)[1] + 1
#                         else:
#                             shift = 1.0
#                         xbaseline += shift + shift / 2.

#                     maxheight = max(maxheight, tree.ntips)                
#                 axes.show = False

#         # None is placeholder until custom mark is finished
#         return self.canvas, axes, None



# class OldCloudTree:
#     """
#     Overlay many tree plots on the same Canvas and Axes. All ToyTrees in the 
#     treelist have already been fixed_order reordered. 
#     """
#     def __init__(self, treelist, **kwargs):

#         # store list of ordered trees
#         self.treelist = treelist

#         # set tip names 
#         if not kwargs.get("tip_labels"):
#             self.tip_labels = self.treelist[0].get_tip_labels()
#         else:
#             self.tip_labels = [
#                 kwargs.get("tip_labels")[name] 
#                 for name in self.treelist[0].get_tip_labels()
#             ]

#         # base style
#         self.ntips = len(self.treelist[0])
#         self.style = TreeStyle('m')
#         self.style.update(kwargs)


#     def update(self, axes=None): 

#         # set reasonable dims if not user entered
#         self.set_dims_from_tree_size()

#         # if not canvas then creates one else uses the existing
#         # sets self.canvas and self.axes
#         self.get_canvas_and_axes(axes)

#         # plot trees on the same axes with shared style dict
#         self.axes.show = False
#         for tre in self.treelist:
#             tre.draw(axes=self.axes, tip_labels=False)

#         # add a single call to tip labels
#         self.add_tip_labels_to_axes()
#         self.fit_tip_labels()

#         # None is placeholder until custom mark is finished        
#         return self.canvas, self.axes, None


#     def get_canvas_and_axes(self, axes):
#         if axes: 
#             self.canvas = None
#             self.axes = axes
#         else:
#             self.canvas = toyplot.Canvas(
#                 height=self.style.height,
#                 width=self.style.width,
#             )
#             self.axes = self.canvas.cartesian(
#                 padding=self.style.padding,
#             )


#     def set_dims_from_tree_size(self):
#         "Calculate reasonable height and width for tree given N tips"
#         tlen = len(self.treelist[0])
#         if self.style.layout in ("r", "l"):
#             # long tip-wise dimension
#             if not self.style.height:
#                 self.style.height = max(275, min(1000, 18 * (tlen)))
#             if not self.style.width:
#                 self.style.width = max(300, min(500, 18 * (tlen)))
#         else:
#             # long tip-wise dimension
#             if not self.style.width:
#                 self.style.width = max(275, min(1000, 18 * (tlen)))
#             if not self.style.height:
#                 self.style.height = max(225, min(500, 18 * (tlen)))


#     def add_tip_labels_to_axes(self):
#         """
#         Add text offset from tips of tree with correction for orientation, 
#         and fixed_order which is usually used in multitree plotting.
#         """
#         # get tip-coords
#         if self.style.layout in ("u", "d"):
#             ypos = np.zeros(self.ntips)
#             xpos = np.arange(self.ntips)

#         elif self.style.layout in ("r", "l"):
#             xpos = np.zeros(self.ntips)
#             ypos = np.arange(self.ntips)

#         else:
#             raise NotImplementedError(
#                 "multitree layout {} not yet supported"
#                 .format(self.style.layout))

#         # pop fill from color dict if using color
#         if self.style.tip_labels_colors:
#             self.style.tip_labels_style.pop("fill")

#         # fill anchor shift if None 
#         # (Toytrees fill this at draw() normally when tip_labels != None)
#         if self.style.tip_labels_style["-toyplot-anchor-shift"] is None:
#             self.style.tip_labels_style["-toyplot-anchor-shift"] = "15px"

#         # add tip names to coordinates calculated above
#         self.axes.text(
#             xpos, 
#             ypos,
#             self.tip_labels,
#             angle=(0 if self.style.layout in ("r", "l") else -90),
#             style=self.style.tip_labels_style,
#             color=self.style.tip_labels_colors,
#         )
#         # get stroke-width for aligned tip-label lines (optional)
#         # copy stroke-width from the edge_style unless user set it
#         if not self.style.edge_align_style.get("stroke-width"):
#             self.style.edge_align_style['stroke-width'] = (
#                 self.style.edge_style['stroke-width'])


#     def fit_tip_labels(self):
#         """
#         Modifies display range to ensure tip labels fit. This is a bit hackish
#         still. The problem is that the 'extents' range of the rendered text
#         is totally correct. So we add a little buffer here. Should add for 
#         user to be able to modify this if needed. If not using edge lengths
#         then need to use unit length for treeheight.
#         """
#         if not self.tip_labels:
#             return 

#         lname = max([len(i) for i in self.tip_labels])

#         # get ratio of names to tree in plot
#         ratio = max(lname / 10, 0.15)

#         # have tree figure make up 85% of plot
#         if self.treelist[0].style.use_edge_lengths:
#             addon = self.treelist[0].treenode.height
#         else:
#             addon = self.treelist[0].treenode.get_farthest_leaf(True)[1] + 1
#         addon *= ratio

#         # modify display for orientations
#         if self.style.tip_labels:
#             if self.style.layout == "r":
#                 self.axes.x.domain.max = addon / 2.
#             elif self.style.layout == "d":
#                 self.axes.y.domain.min = (-1 * addon) / 2

#         # # longest name (this will include html hacks)
#         # longest_name = max([len(i) for i in self.tip_labels])
#         # if longest_name > 10:
#         #     multiplier = 0.85
#         # else:
#         #     multiplier = 0.25

#         # if self.style.use_edge_lengths:
#         #     addon = (self.treelist[0].treenode.height + (
#         #         self.treelist[0].treenode.height * multiplier))
#         # else:
#         #     addon = self.treelist[0].treenode.get_farthest_leaf(True)[1]

#         # # modify display for orientations
#         # if self.style.orient == "right":
#         #     self.axes.x.domain.max = addon
#         # elif self.style.orient == "down":
#         #     self.axes.y.domain.min = -1 * addon


#     # def assign_tip_labels_and_colors(self):
#     #     "assign tip labels based on user provided kwargs"
#     #     # COLOR
#     #     # tip color overrides tipstyle.fill
#     #     # if self.style.tip_labels_color:

#     #     #     if self.fixed_order:
#     #     #         if isinstance(self.style.tip_labels_color, (list, np.ndarray)):
#     #     #             cols = np.array(self.style.tip_labels_color)
#     #     #             orde = cols[self.fixed_idx]

#     #     #     if self.style.tip_labels_style.fill:
#     #     #         self.style.tip_labels_style.fill = None

#     #     # LABELS
#     #     # False == hide tip labels
#     #     if self.style.tip_labels is False:
#     #         self.style.tip_labels_style["-toyplot-anchor-shift"] = "0px"
#     #         self.tip_labels = [
#     #             "" for i in self.mtree._cons_order]

#     #     # (True or None) == use user or cons order
#     #     if self.style.tip_labels is True:
#     #         if self.mtree._user_order:
#     #             self.tip_labels = self.mtree._user_order
#     #         else:
#     #             self.tip_labels = self.mtree._cons_order

#     #     # LABELS
#     #     # user entered something...
#     #     elif isinstance(self.style.tip_labels, list):
#     #         # if user did not change label-offset then shift it here
#     #         if not self.style.tip_labels_style["-toyplot-anchor-shift"]:
#     #             self.style.tip_labels_style["-toyplot-anchor-shift"] = "15px"
#     #         self.tip_labels = self.style.tip_labels



#!/usr/bin/env python

"""Axis class object definition.

In toyplot the Axis is defined in coordinates.py along with
Cartesian, whereas here we split it into a separate module.
"""

from typing import Mapping, Any, List, Sequence, Union, Tuple
from dataclasses import dataclass, field
from collections import defaultdict
import itertools
import numpy
import toyplot


@dataclass(repr=False)
class Helper:
    """Base class for Axis Helpers with a clean repr."""
    def __repr__(self):
        """Return dict-like repr of _{key} attributes in object"""
        data = {i[1:]: j for (i, j) in self.__dict__.items() if i[0] == "_"}
        x = ", ".join(f"{i}={j}" for i, j in data.items())
        return f"{self.__class__.__name__}({x})"


@dataclass(repr=False)
class Label(Helper):
    _text: str = None

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        self._text = value


@dataclass(repr=False)
class Show(Helper):
    _show: bool = True

    @property
    def show(self) -> bool:
        """Control whether the domain should be made visible using the
        axis spine.
        """
        return self._show

    @show.setter
    def show(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError("A boolean value is required.")
        self._show = value


@dataclass(repr=False)
class Near(Helper):
    _near: float = None

    @property
    def near(self) -> float:
        return self._near

    @near.setter
    def near(self, value: float) -> None:
        if value is None:
            self._near = None
        else:
            self._near = toyplot.units.convert(value, target="px", default="px")


@dataclass(repr=False)
class Far(Helper):
    _far: float = None

    @property
    def far(self) -> float:
        return self._far

    @far.setter
    def near(self, value: float) -> None:
        if value is None:
            self._far = None
        else:
            self._far = toyplot.units.convert(value, target="px", default="px")


@dataclass(repr=False)
class TextStyle(Helper):
    """Dictionary of CSS property-value pairs.

    Use the *style* property to control the text appearance.  The
    following CSS properties are allowed:

    * alignment-baseline
    * baseline-shift
    * fill
    * fill-opacity
    * font-size
    * font-weight
    * opacity
    * stroke
    * stroke-opacity
    * stroke-width
    * text-anchor
    * -toyplot-anchor-shift

    Note that when assigning to the *style* property, the properties you
    supply are merged with the existing properties.
    """
    _style: Mapping[str, Any] = field(default_factory=dict)

    @property
    def style(self) -> Mapping[str, Any]:
        return self._style

    @style.setter
    def style(self, value: Mapping) -> None:
        self._style = toyplot.style.combine(
            self._style,
            toyplot.style.require(
                value,
                allowed=toyplot.style.allowed.text,
            )
        )


@dataclass(repr=False)
class Offset(Helper):
    _offset: bool = True

    @property
    def offset(self) -> float:
        return self._offset

    @offset.setter
    def offset(self, value: float) -> None:
        if value is None:
            self._offset = None
        else:
            self._offset = toyplot.units.convert(value, target="px", default="px")


@dataclass(repr=False)
class Location(Helper):
    """Controls the position of ticks (and labels) relative to the axis.

    Allowed values are "above" (force labels to appear above the axis),
    "below" (the opposite), or `None` (use default, context-sensitive
    behavior).
    """
    _location: str = None

    @property
    def location(self) -> str:
        return self._location

    @location.setter
    def location(self, value: str) -> None:
        if value is None:
            self._location = None
        else:
            self._location = toyplot.require.value_in(
                value,
                [None, "above", "below"],
            )


@dataclass(repr=False)
class LineStyle(Helper):
    """Dictionary of CSS property-value pairs.

    Use the *style* property to control the appearance of the line. The
    following CSS properties are allowed:

    * opacity
    * stroke
    * stroke-dasharray
    * stroke-opacity
    * stroke-width

    Note that when assigning to the *style* property, the properties you
    supply are merged with the existing properties.
    """
    _style: Mapping[str, Any] = None

    @property
    def style(self) -> Mapping[str, Any]:
        return self._style

    @style.setter
    def style(self, value: Mapping) -> None:
        self._style = toyplot.style.combine(
            self._style,
            toyplot.style.require(
                value,
                allowed=toyplot.style.allowed.line,
            )
        )


@dataclass(repr=False)
class MinMax(Helper):
    _min: float = None
    _max: float = None

    @property
    def min(self) -> float:
        """Specify an explicit domain minimum for this axis.  By default
        the implicit domain minimum is computed from visible data.
        """
        return self._min

    @property
    def max(self) -> float:
        """Specify an explicit domain maximum for this axis.  By default
        the implicit domain maximum is computed from visible data.
        """
        return self._max

    @min.setter
    def min(self, value: float) -> None:
        self._min = value

    @max.setter
    def max(self, value: float) -> None:
        self._max = value


@dataclass(repr=False)
class DomainHelper(Show, MinMax, Helper):
    """Controls domain related behavior for this axis."""

    # override show.getter
    @property
    def show(self) -> bool:
        """Control whether the domain should be made visible using the
        axis spine.
        """
        return self._show

    # override show.setter
    @show.setter
    def show(self, value):
        toyplot.log.warning("Altering <axis>.domain.show is experimental.")
        self._show = True if value else False


@dataclass(repr=False)
class InteractiveCoordinatesLabelHelper(Show, TextStyle, Helper):
    """Controls the appearance and behavior of interactive coordinate labels."""
    def __post_init__(self):
        self.style = {
            "fill": "slategray",
            "font-size": "10px",
            "font-weight": "normal",
            "stroke": "none",
            "text-anchor": "middle",
        }


@dataclass(repr=False)
class InteractiveCoordinatesTickHelper(Show, LineStyle, Helper):
    """Controls the appearance and behavior of interactive coordinate ticks."""
    def __post_init__(self):
        self.style = {
            "stroke": "slategray",
            "stroke-width": 1.0,
        }


@dataclass(repr=False)
class InteractiveCoordinatesHelper(Show, Location, Helper):
    _label: InteractiveCoordinatesLabelHelper = (
        field(default_factory=InteractiveCoordinatesLabelHelper))
    _tick: InteractiveCoordinatesTickHelper = (
        field(default_factory=InteractiveCoordinatesTickHelper))

    @property
    def label(self) -> InteractiveCoordinatesLabelHelper:
        return self._label

    @property
    def tick(self) -> InteractiveCoordinatesTickHelper:
        return self._tick


@dataclass(repr=False)
class InteractiveHelper(Helper):
    """Controls interactive behavior for this axis."""
    _coordinates: InteractiveCoordinatesHelper = (
        field(default_factory=InteractiveCoordinatesHelper))

    @property
    def coordinates(self) -> InteractiveCoordinatesHelper:
        return self._coordinates


@dataclass(repr=False)
class LabelHelper(TextStyle, Offset, Label, Location, Helper):
    """Controls the appearance and behavior of an axis label."""
    def __post_init__(self):
        self.style = {
            "font-size": "12px",
            "font-weight": "bold",
            "stroke": "none",
            "text-anchor": "middle",
        }


@dataclass(repr=False)
class SpineHelper(Show, LineStyle, Helper):
    """Controls the appearance and behavior of an axis label."""
    _position: str = "low"

    @property
    def position(self) -> str:
        return self._position

    @position.setter
    def position(self, value) -> None:
        """TODO: which other values are allowed or not?"""
        self._position = value


@dataclass(repr=True)
class TickProxy(Helper):
    _tick: Mapping = None
    _allowed: Mapping[str, Any] = None

    @property
    def style(self) -> Mapping[str, Any]:
        return self._tick.get("style", {})

    @style.setter
    def style(self, value: Mapping[str, Any]) -> None:
        self._tick["style"] = toyplot.style.combine(
            self.style,
            toyplot.style.require(value, allowed=self._allowed)
        )


@dataclass(repr=False)
class PerTickHelper(Helper):
    """Controls the appearanace and behavior of individual axis ticks."""
    _allowed: Mapping[str, Any] = field(default_factory=dict)
    _indices: Mapping = field(default_factory=defaultdict)
    _values: Mapping = field(default_factory=defaultdict)

    def __call__(self, index: int = None, value: float = None):
        if index is None and value is None:
            raise ValueError("Must specify tick index or value.")  # pragma: no cover
        if index is not None and value is not None:
            raise ValueError("Must specify either index or value, not both.")  # pragma: no cover
        if index is not None:
            return TickProxy(self._indices[index], self._allowed)
        elif value is not None:
            return TickProxy(self._values[value], self._allowed)

    def styles(self, values: Sequence[float]) -> List[Mapping]:
        """Return styles associated with tick values."""
        # get list of style dicts, or None, in indices order
        results = [
            self._indices[index].get("style", None) if index in self._indices
            else None for index in range(len(values))
        ]
        # get difference between 
        for value in self._values:
            deltas = numpy.abs(values - value)
            results[numpy.argmin(deltas)] = (
                self._values[value].get("style", None))
        return results


@dataclass(repr=False)
class TickLabelsHelper(Offset, Show, TextStyle, Helper):
    """Controls the appearance and behavior of axis tick labels."""
    _angle: float = 0
    _label: PerTickHelper = None

    def __post_init__(self):
        self._label = PerTickHelper(toyplot.style.allowed.text)
        self.style = {
            "font-size": "10px",
            "font-weight": "normal",
            "stroke": "none",
        }

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, value: float) -> None:
        self._angle = value

    @property
    def label(self):
        return self._label


@dataclass(repr=False)
class TicksHelper(Near, Far, Location, Show, LineStyle, Helper):
    """Controls the appearance and behavior of axis ticks."""
    _locator: toyplot.locator.TickLocator = None
    _angle: float = 0.

    _labels: TickLabelsHelper = None
    _tick: PerTickHelper = None

    def __post_init__(self):
        self.show = False
        self._tick = PerTickHelper(_allowed=toyplot.style.allowed.line)
        self._labels = TickLabelsHelper(_angle=self._angle)

    @property
    def labels(self) -> TickLabelsHelper:
        return self._labels

    @property
    def locator(self) -> toyplot.locator.TickLocator:
        return self._locator

    @locator.setter
    def locator(self, value: toyplot.locator.TickLocator) -> None:
        self._locator = value

    @property
    def tick(self) -> PerTickHelper:
        return self._tick


class Axis:
    """Axis for data projection and styling of ticks, labels, and spine.

    """
    def __init__(
        self,
        label: str = None,
        domain_min: float = None,
        domain_max: float = None,
        scale: str = "linear",
        show: bool = True,
        tick_angle: int = 0,
        tick_locator: toyplot.locator.TickLocator = None,
    ):

        self._finalized = None
        self._data_max = None
        self._data_min = None
        self._display_max = None
        self._display_min = None
        self._domain = DomainHelper(_min=domain_min, _max=domain_max)
        self._interactive = InteractiveHelper()
        self._label = LabelHelper()
        self._scale = None
        """: str arg converted to 'linear' or ('log', base) where base is a int."""
        self._show = show
        self._spine = SpineHelper()
        self._tick_labels = []
        self._tick_locations = []
        self._tick_titles = []
        self._ticks = TicksHelper(_tick=tick_locator, _angle=tick_angle)

        self.scale = scale

    @property
    def interactive(self) -> InteractiveHelper:
        return self._interactive

    @property
    def domain(self) -> DomainHelper:
        return self._domain

    @property
    def label(self) -> LabelHelper:
        return self._label

    @property
    def scale(self) -> str:
        return self._scale

    @property
    def show(self) -> bool:
        return self._show

    @property
    def spine(self) -> SpineHelper:
        return self._spine

    @property
    def ticks(self) -> TicksHelper:
        return self._ticks

    @show.setter
    def show(self, value: bool) -> None:
        self._show = value

    @scale.setter
    def scale(self, value: str) -> None:
        if value == "linear":
            self._scale = "linear"
            return
        elif value in ["log", "log10"]:
            self._scale = ("log", 10)
            return
        elif value == "log2":
            self._scale = ("log", 2)
            return
        elif isinstance(value, tuple) and len(value) == 2:
            scale, base = value
            if scale == "log":
                self._scale = ("log", base)
                return
        raise ValueError(
            """Scale must be "linear", "log", "log10", "log2" or a ("log", base) tuple.""")

    def _locator(self) -> toyplot.locator.TickLocator:
        """Return the tick locator set for this Axis, or an appropriate
        one, which can be used to get Tick locations given a domain.

        Exampel: This is called in Cartesian._finalize().
        """
        if self.ticks.locator is not None:
            return self.ticks.locator
        if self.scale == "linear":
            return toyplot.locator.Extended()
        if isinstance(self.scale, toyplot.projection.Projection):
            return toyplot.locator.Null()
        else:
            scale, base = self.scale
            if scale == "log":
                return toyplot.locator.Log(base=base)
        raise RuntimeError("Unable to create an appropriate locator.")  # pragma: no cover

    def _update_domain(self, values, display=True, data=True) -> None:
        """Update axis domain (display and/or data) given a set of
        values from the Marks mapped to Axis.

        This is called in Cartesian._finalize()
        """
        if display:
            self._display_min, self._display_max = toyplot.data.minimax(
                itertools.chain(
                    [self._display_min, self._display_max],
                    values
                )
            )
        if data:
            self._data_min, self._data_max = toyplot.data.minimax(
                itertools.chain(
                    [self._data_min, self._data_max],
                    values,
                )
            )

    def _finalize(
        self,
        x1: float,
        x2: float,
        y1: float,
        y2: float,
        offset: float,
        domain_min: float,
        domain_max: float,
        tick_locations: Sequence[float],
        tick_labels: Sequence[str],
        tick_titles: Sequence[str],
        default_tick_location,
        default_ticks_near,
        default_ticks_far,
        default_label_location,
    ):
        """Project position of spine, ticks and labels given the domain
        and range of data along the Axis.

        Example: This is called at the end of Cartesian._finalize().
        """
        # return if already finalized
        if self._finalized is not None:
            return self._finalized

        # store values to self
        self._x1 = x1
        """: xmin_range (canvas units)."""
        self._x2 = x2
        """: xmax_range (canvas units)."""
        self._y1 = y1
        """: ymin_range (canvas units)."""
        self._y2 = y2
        """: ymax_range (canvas units)."""
        self._offset = offset
        """: padding between axis and data Marks."""
        self._domain_min = domain_min
        """: domain_min (data units)."""
        self._domain_max = domain_max
        """: domain_max (data units)."""
        self._tick_locations = tick_locations
        """: Sequence of locations in data units."""
        self._tick_labels = tick_labels
        """: Sequence of str labels for ticks."""
        self._tick_titles = tick_titles
        """: Sequence of str titles for ticks."""
        self._tick_location = (
            default_tick_location if self.ticks.location is None
            else self.ticks.location
        )
        """: below or above."""
        self._ticks_near = (
            default_ticks_near if self.ticks.near is None
            else self.ticks.near
        )
        """: Length of tick mark inside axis in px units."""
        self._ticks_far = (
            default_ticks_far if self.ticks.far is None
            else self.ticks.far
        )
        """: Length of tick outside axis in px units."""
        self._tick_labels_location = self._tick_location
        """: below or above."""
        self._tick_labels_offset = (
            6 if self.ticks.labels.offset is None
            else self.ticks.labels.offset
        )
        """: padding between tick label and axis in px units."""
        self._label_location = (
            default_label_location if self.label.location is None
            else self.label.location
        )
        """: below or above."""
        self._label_offset = (
            22 if self.label.offset is None else self.label.offset
        )
        """: padding between axis label and axis in px units."""
        self._interactive_coordinates_location = (
            _opposite_location(self._tick_labels_location) if
            self.interactive.coordinates.location is None
            else self.interactive.coordinates.location
        )
        """: below or above."""

        # rowstack [min,min], [max, max]
        endpoints = numpy.row_stack(((x1, y1), (x2, y2)))
        # get matrix norm of [xlen, ylen]
        length = numpy.linalg.norm(endpoints[1] - endpoints[0])
        # get projection of domain_min to (0, length)
        self.projection = _create_projection(
            scale=self.scale,
            domain_min=domain_min,
            domain_max=domain_max,
            range_min=0.0,
            range_max=length,
        )

        self._finalized = self
        return self._finalized


def _create_projection(
    scale: Union[str, Tuple[str, int]],
    domain_min: float,
    domain_max: float,
    range_min: float,
    range_max: float,
) -> toyplot.projection.Projection:
    """Return a Projection to map data units to canvas units given scale type

    ...
    """
    # return the supplied projection
    if isinstance(scale, toyplot.projection.Projection):
        return scale

    # return a linear projection
    args = (domain_min, domain_max, range_min, range_max)
    if scale == "linear":
        return toyplot.projection.linear(*args)

    # return a log projection. log scale format: ('log', int)
    scale, base = scale
    return toyplot.projection.log(base, *args)


def _opposite_location(location: str) -> str:
    """Return the opposite of current location arg."""
    return "above" if location == "below" else "below"


if __name__ == "__main__":

    ax = Axis()
    print(ax)

    ax._finalize(
        x1=0,
        x2=500,
        y1=0,
        y2=0,
        offset=0,
        domain_min=0,
        domain_max=1,
        tick_locations=[0, 0.5, 1],
        tick_labels=['0', '0.5', '1'],
        tick_titles=[],
        default_tick_location="below",
        default_ticks_far=5,
        default_ticks_near=0,
        default_label_location="below",
    )
    print(ax.projection([0, 1]))